# coding: utf-8

"""
Functions for preprocessing of newview data, moved here from POOL\coatings\newview 2018/10/31.

2018/10/29 add rotation of data (optional)

2018/05/11 this script has the only purpose of preprocessing data by removing
    instrument signature and leveling data. All the remaining part must be moved to
    different modules.

    To launch it, create a list of filenames and a single calibration file f2, launch as: calibrate_samples(files,f2,cr=None )
    Return a dictionary with base filename as key and the calibrated data as value.
    Leveling options have been added as parameters.

    Color scale for difference (calibrated data) can be set with cr."""

"""display_diff and calibrate_samples were contained in newview_plotter, they respectively plot differences
and calibrate files.

plot_diffs and get_datasets were in  and roughly a variation of the above. Moved here to try to integrate,
while newview_plotter is removed as the only other thing done is plotting comparison of PSDs that belongs to PSD2D. """

import os
import copy as cp

from pySurf.data2D_class import Data2D,PSD2D
from pySurf.readers.instrumentReader import matrixZygo_reader
from plotting.backends import maximize
from plotting.fignumber import fignumber
#from pySurf.psd2d import rms_power,plot_rms_power

import matplotlib.pyplot as plt
import numpy as np
from pyProfile.profile import polyfit_profile
from IPython.display import display
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.span import span, span_from_pixels
from dataIO.outliers import remove_outliers
import logging
import os
import pdb
from scipy.ndimage import map_coordinates
from scipy import interpolate as ip
from plotting.captions import legendbox
from astropy.io import fits
from plotting.backends import maximize

from scripts.plot_surface_analysis import plot_surface_analysis



def display_diff(f1,f2,data=False,dis=True,crd=None):
    """calculates, plots and returns the difference of two data.
      It is just a diff-and-plot with a few added options and settings for color scale and reading files,
      can probably be replaced by more standard routines.

      f1 and f2 are file names, unless data is set, in that case they are interpreted as data.
      Data are plotted on common color scale, while range for difference crd can be imposed,
      otherwise is calculated on outlier removed difference."""

    #set units
    units=['mm','mm','nm']
    zscale=1000

    #load data in d1 and d2 according if arguments are data or filenames
    if not(data):
        d1=Data2D(file=f1,units=units,scale=(1000,-1000,zscale),center=(0,0))
        if f2 is not None:
            d2=Data2D(file=f2,units=units,scale=(1000,-1000,zscale),center=(0,0))
        else:
            d2=cp.deepcopy(d1)*0
            #d2.data=d2.data*0
    else:
        d1=Data2D(f1.data,f1.x,f1.y,units=units)
        if f2 is not None:
            d2=Data2D(f2.data,f2.x,f2.y,units=units)
        else:
            d2=cp.deepcopy(d1)*0
            #d2.data=d2.data*0

    #plane level and diff
    d1=d1.level()
    d2=d2.level()
    diff=(d1-d2).level()

    #calculate common color range
    cr1=remove_outliers(d1.data,span=True)
    cr2=remove_outliers(d2.data,span=True)
    cr=[min([cr1[0],cr2[0]]),max([cr1[0],cr2[1]])]
    if crd is None: #color range for difference.
        crd=remove_outliers(diff.data,span=True)

    #plot on three panels, prints stats
    plt.clf()
    plt.subplot(311)
    maximize()
    d1.plot()
    plt.clim(*cr)
    d1.printstats()
    plt.subplot(312)
    d2.plot()
    plt.clim(*cr)
    d2.printstats()
    plt.subplot(313)
    diff.plot()
    plt.clim(*crd)
    print ("diff 2 - 1 = ")
    diff.printstats()
    plt.tight_layout()
    #plt.clim(*cr)#[-0.0001,0.0001])
    if not(data):
        if f2 is not None:
            plt.title ('Difference %s - %s'%tuple(os.path.basename(ff) for ff in [f1,f2]))
        else:
            plt.title ('%s '%f2)
    plt.show()

    #print ("PV: ",span(diff.data,size=True),diff.units[-1])
    #print ("rms: ",np.nanstd(diff.data),diff.units[-1])
    return d1,d2,diff

def plot_diffs(dataset,repeatlist,labels=None,outfolder=None,nsigma_crange=1,fignum=None,figsize=(9.6,6.7)):
    """Generate comparison and difference plots for a list of datasets (data,x,y) and a list of indices
    `repeatlist` as (index1,index2,label). An optional list of `labels` matching data in repeatlist can be provided,
    and then are used for titles of plots, otherwise numeric indices are simply used."""

    stats=[]
    result=[]
    #plot surface map, who knows why on figure 5
    fignumber(fignum,figsize=figsize)
    for i1,i2,t in repeatlist:
        d1,d2=dataset[i1],dataset[i2]
        diff=d1[0]-d2[0]
        result.append((diff,d1[1],d1[2]))
        #mng = plt.get_current_fig_manager()
        #mng.window.showMaximized()
        plt.clf()
        ax1,ax2,ax3=diff_images(d1[0],d2[0],d2[1],d2[2],fignum=0)  #x and y are taken from second data
        if labels is not None:
            ax1.set_title(labels[i1])
            ax2.set_title(labels[i2])
            stats.append([labels[i1],labels[i2],np.nanstd(diff)])
        else:
            stats.append([i1,i2,np.nanstd(diff)])
        ax3.set_title('Diff, rms= %5.3f nm'%(np.nanstd(diff)*1000) )
        plt.sca(ax3)
        plt.clim(*filtered_span(diff,nsigma=nsigma_crange,itmax=1,span=True))
        ax1.set_aspect('equal')
        ax2.set_aspect('equal')
        ax3.set_aspect('equal')
        plt.suptitle(t)
        plt.tight_layout()
        if outfolder:
            #pdb.set_trace()
            plt.savefig(os.path.join(outfolder,fn_add_subfix("diff","_%i_%i"%(i1,i2),'.jpg')))
            save_data(os.path.join(outfolder,fn_add_subfix("diff","_%i_%i"%(i1,i2),'.dat')),
                diff,d1[1],d1[2])

    if outfolder:
        np.savetxt(os.path.join(outfolder,"diff_report.txt"),result,fmt="%s")

    return result



def calibrate_samples(files,f2,cr=None,rotate=False,level=((1,1),False)):
    """create a dictionary ddic with calibrated and shape leveled Data2D for each file, key is the base filename.
    Uses display_diff.

    Level is a two element tuple containing (leveling_order,byline) see data2D.level_data."""

    ddic={}
    for ff in files:
        #remove signature
        print("calibrating ",os.path.basename(ff))
        d1,d2,diff=display_diff(ff,f2,dis=False,crd=cr)

        #level and rotate
        diff=diff.level(*level)
        if rotate:
            diff=diff.rot90()

        #plot and saves
        plt.clf()
        lrange=remove_outliers(diff.data,nsigma=3,itmax=1,span=True)
        diff.plot()
        plt.title(d1.name+' leveled' + (' rotated' if rotate else ""))
        plt.clim(*lrange)
        display(plt.gcf())
        diff.printstats(os.path.basename(ff))

        #add data to return dic
        ddic[os.path.basename(ff)]=diff
    return ddic

def getdataset(datalist,outfolder=None,fignum=None,figsize=(9.6,6.7),levelingfunc=None,
    caldata=None,psdrange=[1e-15,1e-6]):
    """for each file in a list read data and make a preview plot using `plot_surface_analysis`,
        return a list of (data,x,y). caldata are optional calibration data of same shape as data."""

    """moved here from newview_plotter. This is a similar function to calibrate samples, """

    if levelingfunc==None:
        levelingfunc= lambda x: x
    result=[]
    if outfolder is not None:
        os.makedirs(outfolder,exist_ok=True)
    fig=fignumber(fignum)
    for ff in datalist:
        plt.clf()
        wdata,x,y=matrixZygo_reader(ff,scale=(1000.,1000,1.),center=(0,0))
        if caldata is not None: wdata=wdata-caldata
        wdata,x,y=levelingfunc(wdata),x,y  #simple way to remove plane
        plot_surface_analysis(wdata,x,y,label=os.path.basename(ff),outfolder=outfolder,nsigma_crange=1,
            fignum=(fignum+1 if fignum else None),psdrange=psdrange,levelingfunc=levelingfunc,frange=[4,100])  #[1e-11,1e-6]
        maximize()
        result.append((wdata,x,y))
    return result



import numpy as np
import matplotlib.pyplot as plt
import utilities.imaging.man as man
import utilities.imaging.fitting as fit
import scipy.ndimage as nd
from linecache import getline
import astropy.io.fits as pyfits
import pdb

def readCylScript(fn,rotate=np.linspace(.75,1.5,50),interp=None):
    """
    Load in data from 4D measurement of cylindrical mirror.
    File is assumed to have been saved with Ryan's scripting function.
    Scale to microns, remove misalignments,
    strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    """
    #Read in values from header
    f = open(fn+'.hdr','r')
    l = f.readlines()
    f.close()
    #Wavelength should never change
    wave = float(l[0].split()[0])*.001 #in microns
    #Ensure wedge factor is 0.5
    wedge = float(l[1])
    if wedge!=0.5:
        print ('Wedge factor is ' + str(wedge))
        pdb.set_trace()
    #Get pixel scale size
    dx = float(l[-1])

    #Remove NaNs and rescale
    d = np.fromfile(fn+'.bin',dtype=np.float32)
    try:
        d = d.reshape((1002,981))
    except:
        d = d.reshape((1003,982))
    d[d>1e10] = np.nan
    d = man.stripnans(d)
    d = d *wave
    d = d - np.nanmean(d)

    #Remove cylindrical misalignment terms
    d = d - fit.fitCylMisalign(d)[0]

    #Rotate out CGH roll misalignment?
    if rotate is not None:
        b = [np.sum(np.isnan(\
            man.stripnans(\
                nd.rotate(d,a,order=1,cval=np.nan)))) for a in rotate]
        d = man.stripnans(\
            nd.rotate(d,rotate[np.argmin(b)],order=1,cval=np.nan))

    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,dx

def readCyl4D(fn,rotate=np.linspace(.75,1.5,50),interp=None):
    """
    Load in data from 4D measurement of cylindrical mirror.
    Scale to microns, remove misalignments,
    strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    """
    #Get xpix value in mm
    l = getline(fn,9)
    dx = float(l.split()[1])*1000.
    
    #Remove NaNs and rescale
    d = np.genfromtxt(fn,skip_header=12,delimiter=',')
    d = man.stripnans(d)
    d = d *.6328
    d = d - np.nanmean(d)

    #d = np.rot90(d,k = 2)
    #Remove cylindrical misalignment terms
    d = d - fit.fitCylMisalign(d)[0]

    #Rotate out CGH roll misalignment?
    if rotate is not None:
        b = [np.sum(np.isnan(\
            man.stripnans(\
                nd.rotate(d,a,order=1,cval=np.nan)))) for a in rotate]
        d = man.stripnans(\
            nd.rotate(d,rotate[np.argmin(b)],order=1,cval=np.nan))

    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,dx

def readConic4D(fn,rotate=None,interp=None):
    """
    Load in data from 4D measurement of cylindrical mirror.
    Scale to microns, remove misalignments,
    strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    """
    #Get xpix value in mm
    l = getline(fn,9)
    dx = float(l.split()[1])*1000.
    
    #Remove NaNs and rescale
    d = np.genfromtxt(fn,skip_header=12,delimiter=',')
    d = man.stripnans(d)
    d = d *.6328
    d = d - np.nanmean(d)
    
    #d = np.rot90(d,k = 2)
    #Remove cylindrical misalignment terms
    conic_fit = fit.fitConic(d)
    d = d - conic_fit[0]

    #Rotate out CGH roll misalignment?
    if rotate is not None:
        b = [np.sum(np.isnan(\
            man.stripnans(\
                nd.rotate(d,a,order=1,cval=np.nan)))) for a in rotate]
        d = man.stripnans(\
            nd.rotate(d,rotate[np.argmin(b)],order=1,cval=np.nan))

    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,dx,conic_fit[1]

def readFlatScript(fn,interp=None):
    """
    Load in data from 4D measurement of flat mirror.
    File is assumed to have been saved with Ryan's scripting function.
    Scale to microns, remove misalignments,
    strip NaNs.
    Distortion is bump positive looking at surface from 4D.
    Imshow will present distortion in proper orientation as if
    viewing the surface.
    """
    #Read in values from header
    f = open(fn+'.hdr','r')
    l = f.readlines()
    f.close()
    #Wavelength should never change
    wave = float(l[0].split()[0])*.001 #in microns
    #Ensure wedge factor is 0.5
    wedge = float(l[1])
    if wedge!=0.5:
        print ('Wedge factor is ' + str(wedge))
        pdb.set_trace()
    #Get pixel scale size
    dx = float(l[-1])

    #Remove NaNs and rescale
    d = np.fromfile(fn+'.bin',dtype=np.float32)
    try:
        d = d.reshape((1002,981))
    except:
        d = d.reshape((1003,982))
    d[d>1e10] = np.nan
    d = man.stripnans(d)
    d = d *.6328
    d = d - np.nanmean(d)
    d = np.fliplr(d)

    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,dx

def readFlat4D(fn,interp=None):
    """
    Load in data from 4D measurement of flat mirror.
    Scale to microns, remove misalignments,
    strip NaNs.
    Distortion is bump positive looking at surface from 4D.
    Imshow will present distortion in proper orientation as if
    viewing the surface.
    """
    #Get xpix value in mm
    l = getline(fn,9)
    dx = float(l.split()[1])*1000.
    
    #Remove NaNs and rescale
    d = np.genfromtxt(fn,skip_header=12,delimiter=',')
    d = man.stripnans(d)
    d = d *.6328
    d = d - np.nanmean(d)
    d = np.fliplr(d)

    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,dx

def write4DFits(filename,img,dx,dx2=None):
    """
    Write processed 4D data into a FITS file.
    Axial pixel size is given by dx.
    Azimuthal pixel size is given by dx2 - default to none
    """
    hdr = pyfits.Header()
    hdr['DX'] = dx
    hdu = pyfits.PrimaryHDU(data=img,header=hdr)
    hdu.writeto(filename,clobber=True)
    return

def read4DFits(filename):
    """
    Write FITS file of processed 4D data.
    Returns img,dx in list
    """
    dx = pyfits.getval(filename,'DX')
    img = pyfits.getdata(filename)
    return [img,dx]

def readCylWFS(fn,rotate=np.linspace(.75,1.5,50),interp=None):
    """
    Load in data from WFS measurement of cylindrical mirror.
    Assumes that data was processed using processHAS, and loaded into
    a .fits file.
    Scale to microns, remove misalignments,
    strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    """
    #Remove NaNs and rescale
    d = pyfits.getdata(fn)
    d = man.stripnans(d)
    d = d - np.nanmean(d)

    #Remove cylindrical misalignment terms
    d = d - fit.fitCylMisalign(d)[0]
    
    # Negate to make bump positive and rotate to be consistent with looking at the part beamside.
    #d = -d
    d = -np.fliplr(d) #np.rot90(d,k = 2)
    
    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d

def readConicWFS(fn,interp=None):
    """
    Load in data from WFS measurement of cylindrical mirror.
    Assumes that data was processed using processHAS, and loaded into
    a .fits file.
    Scale to microns, remove misalignments,
    strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    Returns the data with best fit conic removed, as well as the
    coefficients in the conic fit.
    """
    #Remove NaNs and rescale
    d = pyfits.getdata(fn)
    d = man.stripnans(d)
    d = d - np.nanmean(d)
    
    # Negate to make bump positive and rotate to be consistent with looking at the part beamside.
    #d = -d
    d = -np.fliplr(d) #np.rot90(d,k = 2)
    
    #Remove cylindrical misalignment terms
    conic_fit = fit.fitConic(d)
    d = d - conic_fit[0]
    
    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d,conic_fit[1]

def readFlatWFS(fn,interp=None):
    """
    Load in data from WFS measurement of flat mirror.
    Assumes that data was processed using processHAS, and loaded into
    a .fits file.
    Scale to microns, strip NaNs.
    If rotate is set to an array of angles, the rotation angle
    which minimizes the number of NaNs in the image after
    stripping perimeter nans is selected.
    Distortion is bump positive looking at concave surface.
    Imshow will present distortion in proper orientation as if
    viewing the concave surface.
    """
    #Remove NaNs and rescale
    d = pyfits.getdata(fn)
    d = man.stripnans(d)
    d = -d
    
    #Interpolate over NaNs
    if interp is not None:
        d = man.nearestNaN(d,method=interp)

    return d

#Read in Zygo ASCII file
def readzygo(filename):
    #Open file
    f = open(filename,'r')

    #Read third line to get intensity shape
    for i in range(3):
        l = f.readline()
    l = l.split(' ')
    iwidth = int(l[2])
    iheight = int(l[3])

    #Read fourth line to get phase shape
    l = f.readline()
    l = l.split(' ')
    pwidth = int(l[2])
    pheight = int(l[3])

    #Read eighth line to get scale factors
    for i in range(4):
        l = f.readline()
    l = l.split(' ')
    scale = float(l[1])
    wave = float(l[2])
    o = float(l[4])
    latscale = float(l[6])

    #Read eleventh line to get phase resolution
    f.readline()
    f.readline()
    l = f.readline()
    l = l.split(' ')
    phaseres = l[0]
    if phaseres is 0:
        phaseres = 4096
    else:
        phaseres = 32768

    #Read through to first '#' to signify intensity
    while (l[0]!='#'):
        l = f.readline()

    #Read intensity array
    #If no intensity, l will be '#' below
    l = f.readline()
    while (l[0]!='#'):
        #Convert to array of floats
        l = np.array(l.split(' '))
        l = l[:-1].astype('float')
        #Merge into intensity array
        try:
            intensity = np.concatenate((intensity,l))
        except:
            intensity = l
        #Read next line
        l = f.readline()

    #Reshape into proper array
    try:
        intensity = np.reshape(intensity,(iheight,iwidth))
    except:
        intensity = np.nan

    #Read phase array
    l = f.readline()
    while (l!=''):
        #Convert to array of floats
        l = np.array(l.split(' '))
        l = l[:-1].astype('float')
        #Merge into intensity array
        try:
            phase = np.concatenate((phase,l))
        except:
            phase = l
        #Read next line
        l = f.readline()

    phase = np.reshape(phase,(pheight,pwidth))
    phase[np.where(phase==phase.max())] = np.nan
    phase = phase*scale*o*wave/phaseres
    f.close()
    print (wave, scale, o, phaseres)

    return intensity, phase, latscale

#Convert Zygo ASCII to easily readable ASCII format
def convertzygo(filename):
    #read in zygo data
    intensity,phase,latscale = readzygo(filename)

    np.savetxt(filename.split('.')[0]+'.txt',phase,header='Lat scale: '+\
            str(latscale)+'\n'+'Units: meters')

def make_extent(data,dx):
    return [-float(np.shape(data)[1])/2*dx,float(np.shape(data)[1])/2*dx,-float(np.shape(data)[0])/2*dx,float(np.shape(data)[0])/2*dx]



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 30 23:42:20 2018
https://stackoverflow.com/questions/2478447/find-largest-rectangle-containing-only-zeros-in-an-n%c3%97n-binary-matrix

@author: Vincenzo
"""

from collections import namedtuple
from operator import mul
import numpy as np
import matplotlib.pyplot as plt

from matplotlib import pyplot as plt
from plotting.plot_positions import plot_poly
from pySurf.points import points_find_hull,points_autoresample
from pySurf.points import plot_points,matrix_to_points2
#

from dataIO.span import span
from scipy.ndimage import label, generate_binary_structure 



def area(size):
    return size[0] * size[1]

def max_rectangle_size(histogram):
    Info = namedtuple('Info', 'start height')
    stack = []
    top = lambda: stack[-1]
    max_size = (0, 0, 0) # height, width and start position of the largest rectangle
    pos = 0 # current position in the histogram
    for pos, height in enumerate(histogram):
        start = pos # position where rectangle starts
        while True:
            if not stack or height > top().height:
                stack.append(Info(start, height)) # push
            elif stack and height < top().height:
                max_size = max(max_size, (top().height, (pos - top().start), top().start), key=area)
                start, _ = stack.pop()
                continue
            break # height == top().height goes here

    pos += 1
    for start, height in stack:
        max_size = max(max_size, (height, (pos - start), start), key=area)

    return max_size

def max_rect(mat, value=0):
    """returns (height, width, left_column, bottom_row) of the largest rectangle 
    containing all `value`'s.

    Example:
    [[0, 0, 0, 0, 0, 0, 0, 0, 3, 2],
     [0, 4, 0, 2, 4, 0, 0, 1, 0, 0],
     [1, 0, 1, 0, 0, 0, 3, 0, 0, 4],
     [0, 0, 0, 0, 4, 2, 0, 0, 0, 0],
     [0, 0, 0, 2, 0, 0, 0, 0, 0, 0],
     [4, 3, 0, 0, 1, 2, 0, 0, 0, 0],
     [3, 0, 0, 0, 2, 0, 0, 0, 0, 4],
     [0, 0, 0, 1, 0, 3, 2, 4, 3, 2],
     [0, 3, 0, 0, 0, 2, 0, 1, 0, 0]]
     gives: (3, 4, 6, 5)
    """
    it = iter(mat)
    hist = [(el==value) for el in next(it, [])]
    max_rect = max_rectangle_size(hist) + (0,)
    for irow,row in enumerate(it):
        hist = [(1+h) if el == value else 0 for h, el in zip(hist, row)]
        max_rect = max(max_rect, max_rectangle_size(hist) + (irow+1,), key=area)
        # irow+1, because we already used one row for initializing max_rect
    return max_rect


def internal_nan_mask(a,x=None,y=None):
    """return a mask with True on internal nans."""
  
    if x is None:
        x=np.arange(np.shape(a)[1])
    if y is None:
        y=np.arange(np.shape(a)[0])    
        
    sx=[np.min(x),np.max(x)]
    sy=[np.min(y),np.max(y)]
     
    # b is 1 where a is nan, 0 othewise
    #used to detect when a region is made of nans
    aisnan=np.where(np.isnan(a),True,False)
    #plt.figure(); plot_data(b,x,y)
    #plt.title('b')

    # s is different integer for each contiguous region, can be  
    s=label(aisnan)[0]
    #plt.figure()
    #plot_data(s)
    
    #m_int mask of internal nan regions of a
    m_int=np.zeros(a.shape).astype(bool)
    for i in range(np.max(s)+1):
        c=s==i 
        if aisnan[c].all():
            #plt.figure()
            #plot_data(c,x,y)
            #for each one find hull and check if it touches borders
            p=matrix_to_points2(c,x,y)
            p=p[p[:,2]!=0,:] #keep only points of current region
            ssx=span(p[:,0]) #ranges of points in region
            ssy=span(p[:,1])
            #print(i,sx,sy,ssx,ssy)
            if (ssx[0] > sx[0] and 
                ssx[1] < sx[1] and 
                ssy[0] > sy[0] and 
                ssy[1] < sy[1]):
                #print(i,'internal')
                import pdb
                #pdb.set_trace()
                m_int=np.logical_or(m_int,c!=0)          
            #print(len(np.where(m_int!=0)[0]))
            #plt.figure()
            #plot_points(p,scatter=1)
            #p=p[~np.isnan(p[:,2]),:]
            #h=points_find_hull(p)
            #plot_poly(h,'bo')
            '''
            if not(((h[0,:] <= min(x)).any() or 
                    (h[0,:] >= max(x)).any() or 
                    (h[1,:] <= min(y)).any() or 
                    (h[1,:] >= max(y))).any()):
                print(i,'internal')
                m_int=np.logical_or(m_int,c==1)    
            '''
    return m_int

def find_internal_rectangle(data,x=None,y=None,continuous=False):
    """ return x and y ranges as ((x0,x1),(y0,y1)) that gives masimum area not containing external nans. 
    if contunuous is True, skip patching the interior nans (bad data surrounded by good data) and crops on the largest area without nans, can be used to speed up the calculation and cut out also holes."""
    
    if continuous:
        #not tested
        ic=max_rect(np.isnan(data))
    else:
        mask=np.zeros(data.shape).astype(bool)
        
        mask = ~np.isnan(data) | internal_nan_mask(data,x,y)
        ic=max_rect(~mask)  #note weird order width-height
        
    
    
    ll=(x[ic[2]],y[ic[3]])
    ur=(x[ic[2]+ic[1]-1],y[ic[3]-ic[0]+1])  #last x and y included
    
    return ((ll[0],ur[0]),(ur[1],ll[1]))
    
    

def test_holes():
    from pySurf.data2D import plot_data
    from pySurf.data2D_class import Data2D
    
    nx,ny=300,200
    a=np.random.random(nx*ny).reshape(ny,nx)
    x=np.arange(nx)*3-300
    y=np.arange(ny)*5-500
    a[:5,:]=np.nan
    a[-15:,:]=np.nan
    a[:,-3:]=np.nan
    a[60:100,:20]=np.nan
    
    a[110:120,160:165]=1
    a[110:120,80:95]=np.nan
    a[20:40,105:125]=np.nan
    
    plt.close('all')
    #plt.subplot(121)
    plt.figure()
    plot_data(a,x,y)
    plt.title('crop all external nans')

    #b=np.where(np.isnan(a),1,a)
    #plt.figure(); plot_data(b,x,y)
    #plt.title('b')
    
    r=find_internal_rectangle(a,x,y)
    
    plot_poly([
    [r[0][0],r[1][0]],
    [r[0][1],r[1][0]],
    [r[0][1],r[1][1]],
    [r[0][0],r[1][1]],
    [r[0][0],r[1][0]]])
        
   # plt.figure()
    
if __name__=="__main__":
    # %%TEST

    
    nx,ny=700,500
    a=np.zeros(nx*ny).reshape(ny,nx)
    x=np.arange(nx)*3-300
    y=np.arange(ny)*5-500
    gd=Data2D(a,x,y,name='test_data')
    
    #gd.data[40:60,150:160]=1
    #gd.data[400:420,300:310]=1
    #gd.data[80:100,420:440]=1

    gd.data[50:60,40:60]=1
    gd.data[300:310,600:620]=1
    gd.data[420:440,80:100]=1
    
    gd.data[50:60,340:360]=1
    gd.data[420:440,330:350]=1    
    gd.data[250:260,40:60]=1
    
    plt.close('all')
    
    plt.figure()
    gd.plot()
    
    ic=max_rect(gd.data)  #note weird order width-height
    
    ll=(gd.x[ic[2]],gd.y[ic[3]])
    ur=(gd.x[ic[2]+ic[1]-1],gd.y[ic[3]-ic[0]+1])  #last x and y included
    
    plot_poly([ll,[ur[0],ll[1]],ur,[ll[0],ur[1]],ll],ls='b')

    plt.show()



import numpy as np
from matplotlib import pyplot as plt


def add_clickable_markers(propertyname='markers',fig=None,ax=None,key=None):
    """Makes pickable the image (image only) in the current figure.
    add a list property to the figure.
    Propertyname doesn't really work yet, but works fine with a single set
    of markers named markers"""
    
    #key='ctrl'
    
    def onclick(event):
        button=event.button
        x=event.xdata
        y=event.ydata
        fig=event.canvas.figure
        ax=event.inaxes

        if button==1: 
            if key is None or event.key==key:
            #if event.key==key:
                l=getattr(fig,propertyname)
                l.append([event.xdata, event.ydata])
                #add point
                ax.set_autoscale_on(False)
                ax.plot(x,y,'or',picker=5,markerfacecolor=None)
        if button==3:
            if key is None or event.key==key:
            #if event.key==key:
                l=getattr(fig,propertyname)
                #print 'list',l
                dist=((np.array(l)-(x,y))**2).sum(axis=1)
                pop=l.pop(np.argmin(dist))
                #np.delete(l,np.argmin(dist),0)
                #print 'clicked:',x,y
                #print 'nearest:', pop
                #dist2=[((np.array(pop)-(a.get_xdata(),a.get_ydata()))**2).sum()  for a in ax.lines]
                dist2=[(np.array((pop[0]-a.get_xdata(),pop[1]-a.get_ydata()))**2).sum()  for a in ax.lines]
                #print dist2
                
                ax.lines.pop(np.argmin(dist2))
        
        plt.draw()
        #print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
        #event.button, event.x, event.y, event.xdata, event.ydata)
    
    if fig is None:
        fig=plt.gcf()

    if ~hasattr(fig,propertyname):
        setattr(fig,propertyname,[])
        
    #ax.imshow(im)
    #ax.autoscale(False)
    cid = fig.canvas.mpl_connect('button_press_event', onclick)
    return fig
    
    
    #to disconnect:
    #fig.canvas.mpl_disconnect(cid)
        
    def on_pick(event):
        # not used. this is an alternative way, it automatically manage the selection of the plotted point.
        import matplotlib
        thisline = event.artist
        if isinstance(thisline,matplotlib.lines.Line2D):
            #code to remove point from markers list
            thisline.remove()
            plt.show()
        
        
if __name__=="__main__":

    def retax():
        """diagnostic routine. return coordinates of all plotted points,
        assuming this are the symbols for the plotted markers."""
        ax=plt.gca()
        return [(a.get_xdata(),a.get_ydata()) for a in ax.lines]


    #from generalPlotting.pickPointList import add_clickable_markers
    imfile=r'test\input_data\IMG_4509_crop.jpg'
    a=plt.imread(imfile)
    plt.clf()
    plt.imshow(a)
    f=add_clickable_markers()
    
    plt.show()
    ax=plt.gca()
    
    print('add points with left button, remove closest with right. retax() prints the points plotted on current axes, f.markers print marker list.')
    


import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as m3d

def fit_3D_plane(points,plot=False):
    """Find best fit plane for points. Points is a Nx3 array.
       If a plane is defined by the equation a*x+b*y+c*z+d=0, then the normal has
       direction [a,b,c]. The plane is returned as a tuple (normal,ctr).
       Note that there is some ambiguity in the orientation of the normal. I believe it is related to the sign of d. Also some mistake in plot of normal. """
    # a plane is a*x+b*y+c*z+d=0 [a,b,c] is the normal.
    # The normal is obtained by svd.
    # Note that probably the results can be obtained by same calculation as
    #   line fit, but taking a different row/column
    ctr = points.mean(axis=0)
    x=points-ctr
    M=np.dot(x.T,x)
    normal=np.linalg.svd(M)[0][:,-1]
    if (plot):
        ax=plt.gca()
        ax.scatter3D(*points.T,color='y')
        plt.xlabel('X')
        plt.ylabel('Y')
        normalArrow=np.vstack([ctr,ctr+normal])
        ax.plot3D(*normalArrow.T)
        """
        xx, yy = np.meshgrid(range(10), range(10))
        span=((points.max(axis=0)-points.min(axis=0))/vv[0]).max()
        
        # calculate corresponding z
        z = (-normal[0] * xx - normal[1] * yy - d) * 1. /normal[2]

        # plot the surface
        plt3d = plt.figure().gca(projection='3d')
        plt3d.plot_surface(xx, yy, z)
        
        """
        #calcolate coordinates of corners of the plane to plot
        xy0=points.min(axis=0)[0:2]
        xy1=points.max(axis=0)[0:2]
        xyvert=np.array([(xy0[0],xy0[1]),(xy0[0],xy1[1]),(xy1[0],xy1[1]),(xy1[0],xy0[1])])
        d=-ctr.dot(normal)
        z = (-normal[0] * xyvert[:,0] - normal[1] * xyvert[:,1] - d)/normal[2] #questo e' cannato.
        p=np.vstack([xyvert.T,z]).T
        vert=[[(list(p[i,:])) for i in range(p.shape[0])]]
        poly=m3d.art3d.Poly3DCollection(vert)
        ax.add_collection3d(poly)
        ax.scatter3D(*p.T,color='r')
        plt.show()   
        
    return  normal,ctr
    """z= a + b*y +c*x
    x,y,zi=points[:,0],points[:,1],points[:,2]
    A = np.column_stack((np.ones(x.size), x, y))
    c, resid,rank,sigma = np.linalg.lstsq(A,zi)
    return c"""



from collections import OrderedDict
import numpy as np
import matplotlib.pyplot as plt


class PickControl:
    def __init__(self, fig):
        self.fig = fig
        self._pickers = []
        self._pickcids = []

    def connect_picks(self):
        for i, picker in enumerate(self._pickers):
            if self._pickcids[i] is None:
                cid = self.fig.canvas.mpl_connect('pick_event', picker)
                self._pickers[i] = cid

    def disconnect_picks(self):
        for i, cid in enumerate(self._pickcids):
            if cid is not None:
                self.fig.canvas.mpl_disconnect(cid)
                self._pickcids[i] = None

    def add_pick_action(self, picker):
        if not callable(picker):
            raise ValueError("Invalid picker. Picker function is not callable")
        if  picker in self._pickers:
            raise ValueError("Picker is already in the list of pickers")
        self._pickers.append(picker)
        cid = self.fig.canvas.mpl_connect('pick_event', picker)
        self._pickcids.append(cid)

class KeymapControl:
    def __init__(self, fig):
        self.fig = fig
        # Deactivate the default keymap
        fig.canvas.mpl_disconnect(fig.canvas.manager.key_press_handler_id)
        self._keymap = OrderedDict()
        # Activate my keymap
        self.connect_keymap()
        self._lastkey = None

    def connect_keymap(self):
        self._keycid = self.fig.canvas.mpl_connect('key_press_event',
                                                   self.keypress)

    def disconnect_keymap(self):
        if self._keycid is not None:
            self.fig.canvas.mpl_disconnect(self._keycid)
            self._keycid = None

    def add_key_action(self, key, description, action_func):
        if not callable(action_func):
            raise ValueError("Invalid key action. Key '%s' Description '%s'"
                             " - action function is not a callable" %
                             (key, description))
        if key in self._keymap:
            raise ValueError("Key '%s' is already in the keymap" % key)
        self._keymap[key] = (description, action_func)

    def keypress(self, event):
        action_tuple = self._keymap.get(event.key, None)
        if action_tuple:
            self._lastkey = event.key
            action_tuple[1]()

    def display_help_menu(self):
        print("Help Menu")
        print("Key         Action")
        print("=========== ============================================")
        for key, (description, _) in list(self._keymap.items()):
            print(("%11s %s" % (key, description)))

class ControlSys(KeymapControl, PickControl):
    def __init__(self, fig, im, data):
        self.fig = fig
        self.im = im
        self.data = data
        self.i = 0
        self._hidekey = None
        self._hidecid = None
        KeymapControl.__init__(self, fig)
        PickControl.__init__(self, fig)

		#self._connect('event', function)

        #self.add_key_action('key', 'Description',function)
        #self.add_pick_action(function)
		
    def _emit(self, event, eventdata):
		#this is called every time you want to raise the custom event
        self.fig.canvas.callbacks.process(event, eventdata)

    def _connect(self, event, callback):
        self.fig.canvas.mpl_connect(event, callback)

    # --- Here are defined functions used as callback above ---
    def function(self):
        pass

    #self.fig.canvas.draw_idle()


fig=plt.figure()
ctrl_sys = ControlSys(fig, im, data, polycolls, stormcells, stormmap)
plt.show()



import numpy as np
import warnings

def ispan(array:np.array,size:bool=False,axis:int=None)->np.array:
    """2018/02/09 as span, but works on indices rather than elements."""

    min=np.nanargmin(array,axis=axis)
    max=np.nanargmax(array,axis=axis)
    if size:
        return max-min
    else:
        return np.dstack([min,max]).squeeze()

def span(array:np.array,
        size:bool=False,
        axis:int=None,
        index:bool=False)->np.array:
    """
    Returns min and max of an array together as a array of couples (or their indices or interval size).
    Args:
        size: If set, return the size of the interval instead of the span.
        index: If set, return indices instead of elements.

    Example:

        pts.shape  #pts is an array of x,y,z coordinates
        Out[206]: (270051, 3)

        s=span(pts,axis=0)   #get coordinate span
        s.shape
        Out[208]: (3, 2)

        s[0]  #span of x axis, works easily for 2d arrays, but see notes
              below for ndarrays.
        Out[209]: array([-34.3813775 ,  34.57937714])

        # separate axis spans:
        xs,ys,zs=span(pts,axis=0)
        xs
        Out[229]: array([-34.3813775 ,  34.57937714])

        xl,yl=span(pts0,axis=0)[:2] #get only x and y spans


    Note that couples [min,max] are accessible on last index.
    This has the effect to have a couple where the value returned by
      min and max would be a single value and the rank of the array is not changed.

    Note that the intent of the function is not to return a result that can be addressed as a 2-element list to give min and max as entire arrays (equal to array.min and array.max).
    For this behavior on ndim arrays, use instead directly np.nanmin and np.nanmax.

    In general, indexing span as list will have same effect it has on input array, and iteratively with the last axis being a couple [min,max] instead of a single value.

    `fit_cylinder` uses this function as:

        # pts.shape
        # Out[138]: (270051, 3)

        s=span(pts,axis=0)
        print ('X: [%6.3f,%6.3f]'%tuple(s[0]))
        print ('Y: [%6.3f,%6.3f]'%tuple(s[1]))
        print ('data range: [%6.3f,%6.3f]'%tuple(s[2]))

    """

#    2019/03 [solved 04/01] for some reason span starts to fail where it was not. Note that max and min
#       return an array that has the dimension specified by axis removed. span must return
#       array of same dimension x 2. However its is debatable if returning the two min, max
#       arrays with dimension 2 in place of the removed axis or as first dimension
#       (or return min and max as list of two elements).
#       The question is if I want to return span as an array of intervals (one for each axis) or as a two element list of min and max array.
#       s=span(pts,axis=0)
#
#       Note that fit_cylinder expects the first option.
#
#   this is solved and updated in documentation.

#   note that this `solved` 19/04 fixes the issue with print as in docstring,
#     but had to change also in points_find_grid:
#
#        #xs,ys,zs=np.hsplit(span(points,axis=0),3)
#        xs,ys,zs=span(points,axis=0)
#
#     and:
#        xl,yl=span(pts0,axis=0)[:2] #was np.transpose(span(pts0,axis=0)[:,:2])
#
#   this was recently failing too.

#    index introduced 2018/02/09
#    axis introduced 2015/12/17"""
    
    # questo risolve errore se si passa array vuoto.
    # Ancora, se passata lista [array([]),array]
    #   da errore poco chiaro (non si dovrebbe fare comunque).
    if np.size(array)==0:
        return array
    
    if index:
        return ispan(array,size=size,axis=axis)
    min=np.nanmin(array,axis=axis)
    max=np.nanmax(array,axis=axis)
    if size:
        return max-min
    else:
        return np.stack([min,max],axis=-1) # couples [min,max] accessible with
        #  instead of scalar elements as in e.g. np.min
        #  see explanation in docstring.
        # it was before 2019/04/01
        # return np.stack([min,max]).squeeze()  #?

def span_from_pixels(p,n=None,delta = 0):
    """From an array with positions of pixel centers `p`, returns the entire range for pixels from side to side. Useful to adjust plot extent in imshow.
    
    `n` is the number of pixels to consider in the output range.
    
    `p` (defaults to n) is uniquely used to extract the range, 
    which makes it possible, as alternative, to pass `p` as range.

    `delta` is used only if p has a single value (this is tested in the simplest way by checking `len(p) == 1`, doesn't account for any complex case, e.g. repeated values).
    
    In this case, the pixel width cannot be evaluated from the single `p` value, and `delta` is assumend as half width of the pixel to use. Can be set to 0 returns two identical values, as default in `dataIO.span()`).
    Not the most elegant, but if a negative value is passed, this is used as fraction of value. 
    
    Usage example:
        
        s = span_from_pixels([100],n,delta=0.1)
        xs = np.linspace(s[0],s[1],n+1) # return the n+1 interval extremes 
    
    More examples in `test_span_from_pixels`.

    note that np.linspace has flag retsteps to return step size."""

    p = np.array(p)
    
    if n is None:
        n=len(p)

    # dx half-width of pixel
    if len(p) == 1:
        #use as fraction if negative
        if delta < 0:
            delta = np.abs(delta)*np.array(p)
        p =  (p-delta,p+delta)
        dx = 0
            
    if n > 1:
        dx = span(p,size=True)/(n-1)
    
    return (span(p)[0]-dx/2,span(p)[1]+dx/2)

def test_span_from_pixels():
    print (span_from_pixels([0,3],4)) #[-0.5,3.5]
    print (span_from_pixels([0,2],3)) #[-0.5,2.5]
    print (span_from_pixels([0,1,2])) #[-0.5,2.5]
    print (span_from_pixels([0,0.5,1,1.5,2])) #[-0.25,2.25]
    #new
    print (span_from_pixels([3],4)) #[3,3]
    print (span_from_pixels([100],1,delta=0.1))  #[99.9, 100.1]
    print (span_from_pixels([100],3,delta=-0.1))  #[85., 115.] 
    print (span_from_pixels([100],2,delta=1))  #(98.0, 102.0)
    print (span_from_pixels([100],3,delta=1))  #(98.5, 101.5)
    
def filtered_span(*args,**kwargs):
    from dataIO.outliers import remove_outliers
    warnings.warn('This routine was replaced by outliers.remove_outliers please update calling code, will be removed.',RuntimeWarning)
    return remove_outliers(*args,**kwargs)

    '''
def filtered_span(  data:np.array,
                    nsigma:float=3,
                    itmax:int=100,
                    span:bool=False,
                    print_partial:bool=False)->np.array:

    """determine the rms after a number of iterations of outlier removals.
        For each iteration, outliers outside `nsigma` standard deviations from average are removed.
        Loop ends after `itmax` iterations or if convergency is reached (two consecutive iterations with same stdev).

    see also dataIO.outliers
        """
    #see also dataIO.outliers
    #data=data.flatten()
    data=data[np.isfinite(data)]
    sigma=np.nanstd(data)
    sigmast=0
    i=1
    while sigma != sigmast:
        sigmast=sigma

        data=data[np.abs(data-np.nanmean(data))<nsigma*sigma]
        sigma=np.nanstd(data)
        i=i+1
        if i >= itmax or sigma==0:
            break
        if print_partial:
            print (sigma,sigmast)
    return np.nanmean(data)+np.array((-1,1))*sigma if span else sigma
    '''



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 06 16:06:48 2016
v1 Saved 2018/08/27 after radical change to raimondi-spiga routine
without saving previous version.
It might be worth to recover it from backups.

@author: Vincenzo Cotroneo
@email: vcotroneo@cfa.harvard.edu
"""
from dataIO.span import span
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats


## profile creation

def line(x,y=None):
    """return line through end points of x,y.
	x and y are vectors, can be of different length (e.g. y can be a 2-elements vector).
    If only one vector is provided, it is assumend as equally spaced points.
"""
    x=np.array(x)
    if y is None:
        y=x
        x=np.arange(len(y))
    y=np.array(y)
    
    #account for nan
    sel=~np.isnan(y)
    if sel.any():
        y0=[y[sel][0],y[sel][-1]]
        if len(x)==len(y):
            x0=x[sel]
        else:
            x0=x
        L=span(x0,size=1)
        return (x-x0[0])*(y0[-1]-y0[0])/L+y0[0]
    else:
        return y
        
def make_signal(amp,L,N,nwaves,ystartend=(0,0),noise=0.,minus_one=False):
    """Build a signal of length L and number of points N, as a sum of a sinusoid, a line and a noise. minus_one remove last point (just a convenience e.g. 
    for a periodic profile)."""
    
    x=np.arange(N,dtype=float)/(N-1)*L
    l=line(x,ystartend)
    y=amp*np.sin(2*np.pi*x/L*nwaves) + noise*np.random.random(N)+l
    if minus_one:
        x,y=x[:-1],y[-1]
    return x,y

def make_circle(x,c,r,sign=1):
    """plot positive part if sign is positive, negative if negative."""
	
    y=np.sqrt(R**2-(x-c[0])**2)+c[1]
    return x,y*np.sign(sign)
    

#profile fitting

def polyfit_profile (x,y=None,degree=1):
    """return polynomial of degree that fits the profile.
"""
    if y is None:
        y=x
        x=np.arange(len(y))
    
    sel=~np.isnan(y)
    if sel.any():
        y0=y[sel]
        x0=x[sel]
        coeff=np.polyfit(x0,y0,degree)
        return np.polyval(coeff,x)
    else:
        return y    

        
##PROFILE I/O
def save_profile(filename,x,y,**kwargs):
    
    np.savetxt(filename,np.vstack([x,y]).T,**kwargs)
        
#PROFILE OPERATIONS        
def level_profile (x,y=None,degree=1):
    """return profile after removal of a polynomial component.
"""
    if y is None:
        y=x
        x=np.arange(len(y))
    
    return x,y-polyfit_profile(x,y,degree) 
        

def remove_profile_outliers(y,nsigma=3,includenan=True):
    """remove outliers from a profile by interpolation.
    y is modified in place. Mask of outliers is returned. """
    
    mask=np.abs(y-np.nanmean(y))<(nsigma*np.nanstd(y)) #points to keep
    if includenan:
        mask[np.isnan(y)]=False
        
    if mask.any()==False:
        return y*np.nan  #all invalid values, transform to nan so it can be
            #detected on call. numeric values cannot be all out of nsigma, unless small.

    if not mask.all():
        x=np.arange(len(y))
        y[mask==False] = np.interp(x[mask==False], x[mask], y[mask])
        
    return mask==False
	

    
"""
def movingaverage(values,window,mode='full'):
    weigths = np.repeat(1.0, window)/window
    #including valid will REQUIRE there to be enough datapoints.
    #for example, if you take out valid, it will start @ point one,
    #not having any prior points, so itll be 1+0+0 = 1 /3 = .3333
    smas = np.convolve(values, weigths, mode)
    return smas # as a numpy array
"""   

def movingaverage(values,window):
    """for now, only odd windows. Full returns"""
    assert int(window/2.)*2==(window-1)
    weigths = np.repeat(1.0, window)/window
    #including valid will REQUIRE there to be enough datapoints.
    #for example, if you take out valid, it will start @ point one,
    #not having any prior points, so itll be 1+0+0 = 1 /3 = .3333
    smas = np.convolve(values, weigths, 'same')
    for i in np.arange(int(window/2.)+1):
        smas[i]=np.mean(values[:2*i+1])
        smas[-i]=np.mean(values[-2*i-1:])
    return smas # as a numpy array

def rebin_profile(x,y,*args,**kwargs):
    ss=stats.binned_statistic(x,y,statistic='mean',*args,**kwargs)
    x2=np.array([(x+y)/2. for x,y in zip(ss[1][:-1],ss[1][1:])])  #centra su punto centrale
    y2=ss[0]
    return x2,y2  

def crop_profile(x,y,xrange=None,*args,**kwargs):
    
    if xrange is None: 
        xrange={None,None}
    if xrange[0] is None:
        xrange[0]=np.min(x)
    if xrange[1] is None:
        xrange[1]=np.max(x)
    
    sel=(x>=xrange[0])&(x<=xrange[1])

    return x[sel],y[sel]  
        
def subtract_profiles(x1,y1,x2,y2,*args,**kwargs):
    return x1,y1-np.interp(x1,x2,y2)  

def calculate_barycenter(x,y):
    """return the x of barycenter using y as weight function."""
    return np.sum(x*y)/len(x)
    
def mirror_profile(xx,yy,center=None,split=False,xout=None):
    """Return the profile x, y mirrored around x=center.
    If split is set return a list of the two profiles >,< 
    as [[xpos,ypos],[xneg,yneg]].
    If xout is passed it is used to interpolate the double
    profiles. If split is not selected and the initial profile
    is not symmetrical can give unexpected results. 
    In case split is selected and center exactly matches one of the input points,
    center point is added to both profiles.
    """
    x=xx.copy()
    y=yy.copy()
    mpos = (x >= center) #mask
    x[~mpos]=2*center-x[~mpos]
    
    if split:
        if center in x:
            mneg=np.logical_not(mpos)
            mneg[x==center]=True
        x0,y0=[x[mneg],y[mneg]],[x[mpos],y[mpos]] #note that x0 and y0 are no more x and y, rather two profiles returned (no sorting)
        if xout is not None:
            x0,y0=[[xout,np.interp(xout,np.sort(xx),yy[xx.argsort()])] for xx,yy in [x0,y0]]
            #x0,y0=[[xout,np.interp(xout,xx,yy)] for xx,yy in [x0,y0]]
    else:
        i0=x.argsort()
        x0=x[i0]
        y0=y[i0]        
        if xout is not None:
            y0=np.interp(xout,x0,y0)
            x0=xout
    
    return x0,y0
        
        
    
def calculate_HEW(x,y,center=None,fraction=0.5):
    """calculate HEW around center from profile x and y by integrating the two sides. If center is None, barycenter is calculated.
    Radius is returned that gives integrated height equal to `fraction`
    of total integral."""
    
    x2 = np.array(x)
    y2 = np.array(y)
    
    if center is None:
        center=calculate_barycenter(x,y)
    
    xe,ye=mirror_profile(x,y,center=center)
    intprof=[np.trapz(ye[:i],xe[:i]) for i in np.arange(len(xe))]
    he=intprof[-1]*fraction
    return xe[np.count_nonzero(intprof<=he)]
    
    
    

#PROFILE STATS AND DERIVED QUANTITIES
def PSF_spizzichino(x,y,alpha=0,xout=None,energy=1.,level=True):
    """Try to use spizzichino theory as in PR notes to calculate Hthe PSF,
    return a vector of same length as xout.
    alpha is incidence angle from normal in degrees, alpha= 90 - shell slope for tilt removed profiles.
        Tilt can be included in the profile, in that case alpha is left to 0
       (total slope must be <0 and alpha>0, this means that profile with tilt is expected to be <0).
    xout can set the output intervals in theta on the focal plane (from specular angle), if not set 512 points are used.
    Lambda is wavelength in keV."""
    
    lambda_mm=12.398425/energy/10**7
    
    if xout is None:
        lout=1001
    else:
        lout=len(xout)
    L=span(x,size=True)
    deltax=L/(len(x))
    
    #calculate and remove slope as alpha. Profile tilt removed is yl
    if level: 
        slope=line(x,y)
        yl=y-slope
        #adjust incidence angle to include the slope removed from profile leveling. 
        # Increasing profile is positive slope angle:
        alpha=alpha*np.pi/180-np.arctan2(y[-1]-y[0],x[-1]-x[0])
    else:
        yl=y
        
    if alpha<=0: raise ValueError
    
    thmax= lambda_mm/(2*deltax*(np.pi/2-alpha))
    #xout is the array of theta for the output
    if xout is None:
        xout=np.linspace(alpha-thmax,alpha+thmax,lout)
    else:
        xout=xout+alpha
        
    scale=np.sqrt(2.)
    I=np.array([np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2 for theta in xout])
    """
    The above is equivalent to (iterate on all theta in xout):
    I[theta]=np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2
    """
    return xout-alpha,I

def PSF_raimondiSR(x,y,alpha=0,xout=None,energy=1.):
    """Try to use theory from Raimondi and Spiga A&A2015 to calculate the PSF for single reflection,
    return a vector of same length as xout.
    alpha is incidence angle from normal in degrees, alpha= 90 - shell slope for tilt removed profiles.
    Tilt can be included in the profile, in that case alpha is left to 0
       (total slope must be <0 and alpha>0, this means that profile with tilt is expected to be <0).
    xout can set the output intervals in theta on the focal plane (from specular angle), if not set 512 points are used.
    Lambda is wavelength in keV.
    """
    
    """
    R0 e' necessario ma solo per normalizzare la PSF.  In realta' se nella formula all'ultima riga sostituisci dr1 = L*sin(alpha) = L*R0/2f (usi la singola riflessione, giusto?) vedrai che R0 se ne va e non serve saperlo. f a questo punto sarà semplicemente la distanza alla quale si valuta il campo, non necessariamente la focale.
    
    dr1 = L*sin(alpha) = L*R0/2f
    
    """
    
    lambda_mm=12.398425/energy/10**7
    
    if xout is None:
        lout=1001
    else:
        lout=len(xout)
    L=span(x,size=True)
    deltax=L/(len(x))
    
    #calculate and remove slope as alpha. Profile after tilt removal is yl
    slope=line(x,y)
    yl=y-slope
    #adjust incidence angle to include the slope removed frorm profile leveling. 
    # Increasing profile is positive slope angle:
    alpha=alpha*np.pi/180-np.arctan2(y[-1]-y[0],x[-1]-x[0]) 
    if alpha<=0: raise ValueError
    
    thmax= lambda_mm/(2*deltax*(np.pi/2-alpha))
    #xout is the array of theta for the output
    if xout is None:
        xout=np.linspace(alpha-thmax,alpha+thmax,lout)
    else:
        xout=xout+alpha
    
    dR1=L*np.sin(alpha)  #L1
    R0=1  #F*np.tan(4*alpha)
    d20=1  #randdomly set things to 1
    z1=1
    
    
    
    PSF=1/2*F/(L*lambda_mm)*np.abs(
    [np.sqrt(y/d20)*np.exp(-2*np.pi*1.j/lambda_mm*(d20-z1+x**2/(2*(S-x))))].sum()*deltax
    )**2  #L1
    
    
    
    scale=np.sqrt(2.)
    I=[np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2 for theta in xout]
    """
    The above is equivalent to (iterate on all theta in xout):
    I[theta]=np.abs((deltax/L*(np.exp(2*np.pi*1.j/lambda_mm*(x*(np.sin(alpha)-np.sin(theta))-scale*yl*(np.cos(alpha)+np.cos(theta)))))).sum())**2
    """
    return xout-alpha,I    
 
'''
#P.S. I suspect this is the same as removing average tilt. Proof??
def autotilt(x,y):
    """Transform a profile by tilting in a way that two contact points are on same horizontal line, return removed line."""
    if delta[0]==0 and delta[-1]==0
    L=span(x,size=1)
    line=x*(delta[-1]+delta[0])/L-delta[0]
    delta=delta-line  #start from endpoints
    i,j=delta.argsort()[:2]   #index of two maxima
'''

##TESTS AND USE

def test_mirror(xx=None,yy=None,center=None,outfolder=None):
    
    print ("mirror a profile about a center coordinate")
    xx=np.arange(30)
    yy=-0.03*xx**2+0.2*xx-5
    center=15
    
    
    plt.close('all')
    plt.figure()
    plt.plot(xx,yy,label='starting profile')
    plt.plot(*mirror_profile(xx,yy,center=center),label='reflect about 15')
    plt.plot(*mirror_profile(xx,yy,center=center),'o')
    for i,p in enumerate(mirror_profile(xx,yy,center=center,split=1)):
        plt.plot(*p,'-.',label='reflect about 15, split#%i'%i) 
    plt.legend(loc=0)

    plt.figure()
    xout=15+np.arange(5)*2
    plt.title('test interpolated output')
    plt.plot(xx,yy,label='starting profile')
    plt.plot(*mirror_profile(xx,yy,center=center,xout=xout),label='reflect about 15')
    plt.plot(*mirror_profile(xx,yy,center=center,xout=xout),'o')
    for i,p in enumerate(mirror_profile(xx,yy,center=center,split=1,xout=xout)):
        plt.plot(*p,'-.',label='reflect about 15, split#%i'%i) 
    plt.legend(loc=0)
    
    plt.show()
    print('done!')
    return xx,yy

def test_HEW():
    #datafile=r'test\01_mandrel3_xscan_20140706.txt'
    print ("uses Spizzichino's formula to predict PSF on sinusoidal ")
    x=np.linspace(0,300.,100)
    y=np.cos(6*np.pi*x/span(x,size=True))/60000.
    xout=np.linspace(-5,5,1000)/206265.
    #x,y=np.genfromtxt(datafile,unpack=True,delimiter=',')
    #y=y-line(x,y)
    plt.figure('profile')
    plt.clf()
    plt.title('profile')
    plt.plot(x,y)
    plt.plot(x,line(x,y))
    plt.xlabel('Axial Position (mm)')
    plt.ylabel('Profile height (mm)')
    plt.figure('PSF')
    plt.clf()
    alpha=89.79
    plt.title('PSF, alpha=%f6.2'%alpha)
    xout,yout=PSF_spizzichino(x,y,alpha=alpha,xout=xout)
    plt.xlabel('angular position around alpha (arcsec)')
    plt.ylabel('Intensity (a.u.)')
    plt.plot(xout*206265.,yout)
    plt.show()
    print('done!')
    return xout,yout
    
if __name__=="__main__":
    test_HEW()
    


"""
Created on Tue Dec 19 14:12:22 2017
https://gist.github.com/g-s-k/ccffb1e84df065a690e554f4b40cfd3a
@author: gkaplan
"""

# %% package dependencies
import sys
import h5py
import numpy as np
import matplotlib.pyplot as plt

# %% loader function
def datx2py(file_name):
    # unpack an h5 group into a dict
    def _group2dict(obj):
        return {k: _decode_h5(v) for k, v in zip(obj.keys(), obj.values())}

    # unpack a numpy structured array into a dict
    def _struct2dict(obj):
        names = obj.dtype.names
        return [dict(zip(names, _decode_h5(record))) for record in obj]

    # decode h5py.File object and all of its elements recursively
    def _decode_h5(obj):
        # group -> dict
        if isinstance(obj, h5py.Group):
            d = _group2dict(obj)
            if len(obj.attrs):
                d['attrs'] = _decode_h5(obj.attrs)
            return d
        # attributes -> dict
        elif isinstance(obj, h5py.AttributeManager):
            return _group2dict(obj)
        # dataset -> numpy array if not empty
        elif isinstance(obj, h5py.Dataset):
            d = {'attrs': _decode_h5(obj.attrs)}
            try:
                d['vals'] = obj[()]
            except (OSError, TypeError):
                pass
            return d
        # numpy array -> unpack if possible
        elif isinstance(obj, np.ndarray):
            if np.issubdtype(obj.dtype, np.number) and obj.shape == (1,):
                return obj[0]
            elif obj.dtype == 'object':
                return _decode_h5([_decode_h5(o) for o in obj])
            elif np.issubdtype(obj.dtype, np.void):
                return _decode_h5(_struct2dict(obj))
            else:
                return obj
        # dimension converter -> dict
        elif isinstance(obj, np.void):
            return _decode_h5([_decode_h5(o) for o in obj])
        # bytes -> str
        elif isinstance(obj, bytes):
            return obj.decode()
        # collection -> unpack if length is 1
        elif isinstance(obj, list) or isinstance(obj, tuple):
            if len(obj) == 1:
                return obj[0]
            else:
                return obj
        # other stuff
        else:
            return obj

    # open the file and decode it
    with h5py.File(file_name, 'r') as f:
        h5data = _decode_h5(f)

    return h5data


# %% script body
if __name__ == '__main__':
    # use function above
    h5data = datx2py(sys.argv[1])
    # parse out height values
    zdata = h5data['Data']['Surface']
    zdata = list(zdata.values())[0]
    zvals = zdata['vals']
    zvals[zvals == zdata['attrs']['No Data']] = np.nan
    # get units
    zunit = zdata['attrs']['Z Converter']['BaseUnit']
    # get stats
    pv = np.nanmax(zvals) - np.nanmin(zvals)
    rms = np.sqrt(np.nanmean((zvals - np.nanmean(zvals))**2))
    # display
    plt.pcolormesh(zvals)
    plt.title("{0:0.3f} {2} PV, {1:0.3f} {2} RMS".format(pv, rms, zunit))
	


from pyProfile.profile import line, make_signal
import numpy as np
from dataIO.span import span
import matplotlib.pyplot as plt
#from pySurf.data2D import projection

def psd2prof(f,p,phase=None,N=None):
    """build a profile from PSD, if phase (in rad) is not passed use random 
    values. A real profile of npoints is assumed. Note that profiles with 
    (len(p)-1)*2 and (len(p)-1)*2+1 points give psd with same number of points,
    so both reconstructions are possible. Odd number of points is the default,
    passing N overrides, with N=(len(p)-1)*2, default (len(p)-1)*2+1"""
 
    if phase==None:
        phase=np.random.random(len(f))*np.pi*2
    
    
    #adjust norm of positive freq.
    p=p/2  
    p[0]=p[0]*2
    #add complex part
    #p[0]=p[0]/2
    yfft=np.sqrt(p)*np.exp(1j*phase)
    if N is None:
        N=(len(p)-1)*2+1
    else:
        if N!=(len(p)-1)*2 and N!=(len(p)-1)*2+1 :
            raise ValueError
    y=np.fft.irfft(yfft,N)
    y=y*len(y)
    x=np.arange(len(y))/(f[1]-f[0])/(len(y)-1)
    x=x-x[0]
    
    return x,y

def normPSD(N,L=None,form=1):
    """return a normalization factor for the PSD, calculated according to different definitions of `power`.
    The PSD is defined as PSD(f)=factor*np.abs(FFT)**2 where FFT is the fast fourier transform.
    Note that in case of rfft the doubling factor is not included in norm (because part of FFT definition).
    Possible values are (ref. to Numerical recipes when possible):
    1 - units: [Y**2] no normalization. "sum squared amplitude" in NR 13.4.1
    2 - units: [Y**2][X] This has the advantage that measurements on same range with different number of points match and it is the way it is usually plotted. The rms is the integral of PSD over frequency rather than the sum.
    3 - units: [Y**2] here rms is the sum of PSD, this is RA normalization (however it differs in RA when a window is used). 13.4.5 in NR.
    4 - units: [Y**2] not sure what this is for, "mean squared amplitude" in NR 13.4.2 and in formula 12.1.10 as discrete form of Parsifal's theorem.
    """
        
    if form==0:
        factor=1. #units: [Y**2] no normalization. "sum squared amplitude" in NR 13.4.1
    if form==1:
        factor=1./N**2*L #[Y**2][X] This has the advantage that measurements on same range with different number of points match
        #and it is the way it is usually plotted. The rms is the integral of PSD over frequency rather than the sum.
    elif form==2:
        factor=1./N**2 #[Y**2] here rms is the sum of PSD, this is RA normalization (however it differs in RA when a window is used).
        #"" 13.4.5 in NR
    elif form==3:
        factor=1./N #[Y**2] not sure what this is for, "mean squared amplitude" in NR 13.4.2
                    #and in formula 12.1.10 as discrete form of Parsifal's theorem
    return factor
    
def psd(x,y,retall=False,wfun=None,norm=1,rmsnorm=False):
    """return frequencies and PSD of a profile.
    
    If retall is set, also the phase is returned.
    PSD is squared FFT divided by step size. Profile is assumed real, so only positive freqs are used, doubling PSD values. Return value has ceil(N/2) elements (always odd).
    See `np.fft.rfftfreq` for details.
    wfun is a function that given the number of points return a vector with values for the window. Note that normalization is quite arbitrary, usually it's rescaled later with rmsnorm.
    2017/07/30 if rmsnorm is set True, PSD is normalized to (multiplied by) rms calculated from profile (must be same if no window applied). 
    2016/03/26 this psd i return to old normalizaiton "time-integral squared magnitude", multiplying by dx.
    """
    
    if wfun is None:
        win=np.ones(len(y))
    else:
        win=wfun(len(x))
    N = len(y)
    L=span(x,True)
    
    yfft  = np.fft.rfft(y*win) #note that rfft return exact values as fft (needs to be doubled after squaring amplitude)
    
    normfactor=normPSD(N,L,form=norm)
    if rmsnorm:  #normalize to rms of non windowed function. This is independent on 0 constant term, 
    # but also alter its meaning 
        normfactor=normfactor*np.nanstd(y)**2/np.nanstd(y*win)**2
   
    psd  = 2*normfactor*np.abs(yfft)**2
    psd[0]=psd[0]/2  #the first component is not doubled
    
    freqs=np.fft.rfftfreq(N,np.float(L)/(N-1))
    if retall:
        return freqs,psd,np.angle(yfft)
    else:
        return freqs,psd         
        
def psd_units(units=[None,None,None]):
    """return as a list of 3 strings (or None) for `units` of x,f,PSD
    from units of x,y,data.
    In case there is not enough information on units of data, 
        returned units are '', '[Y]', '[Y] [Z]$^2$'.
    
    This aims to keep some consistency in handling units, because
    of the risk of ambiguity, e.g. m^2 vs m**2 or m$^2$,
    or even None vs "".
    This ambiguity is also the reason why in object version `Data2D_class.PS2D` 
    units are kept from surface data, rather than being set to PSD units.
    This is probably not the best way to solve the ambiguity, but this
    function should be the preferred way to generate units for the 
    PSD and its axis, until a better one is found. 
    X is left blank rather than outputting '[X]' because it is identical
        to input units and it can be inconvenient to use the string
        in plots (e.g. "x ([X])") and the original units can be accessed
        if desired.
        
    2020/07/16 moved to `pyProfile.psd` from `pySurf.psd2d`, making it valid for 2 or 3D units."""
    
    import copy
    if units is not None:
        units = units.copy() #otherwise bizarre side effect
    flag2d = False  #flag to return 2-el vector
    if np.size(units) == 1: 
        units = np.repeat(units,3)
    elif np.size(units) == 2:
        flag2d=True
        units=[None,units[0],units[1]]

    units[1] = (units[1] if units[1] else "[Y]")+"$^{-1}$"
    if units[0] is None: units[0] = "" # cbunits[0] = units[0] if units[0] else ""
    if units[2] and units[1]:
        units[2] = units[1]+" "+units[2]+"$^2$"  #colorbar units string
    else:
        units[2]=(units[1] if units[1] else "[Y]")+" [Z]$^2$"
        #units[2]="[Y] [Z]$^2$"
    
    if flag2d: units=units[1:3]
    return units
    
def test_psd_units():
    """run tests with different combinations of defined and undefined units.
    """
    
    testval = [[None,None],
               ['mm',None],
               ['mm',''],
               ['mm','mm'], #the following fails:
               None,  
               'mm',
               ['mm']]
    
    print ('TEST FOR 2D UNITS (PROFILE)')
    for t in testval:
        print ("test for: ",t)
        print ('PSD units(x,f,PSD):\n'+'\n'.join(psd_units(t)))
        print ('----------\n')

    testval = [[None,None,None],
               ['mm',None,None],
               ['mm',None,''],
               ['mm','',''],
               ['mm','mm',''],
               ['mm','mm','mm'],  #the following fails:
               None,  
               'mm',
               ['mm']]
               
    print ('TEST FOR 3D UNITS (SURFACE)')
    for t in testval:
        print ("test for: ",t)
        print ('PSD units(x,f,PSD):\n'+'\n'.join(psd_units(t)))
        print ('----------\n')

def components(d,win=1):
    """RA Want to return Fourier components with optional window
    Application note: These components are dependent on sampling!
    This means you can *not* interpolate these components onto other
    frequency grids!
    """
    #Handle window
    if win != 1:
        if np.size(np.shape(d)) == 1:
            win = win(np.size(d))/np.sqrt(np.mean(win(np.size(d))**2))
        else:
            win1 = win(np.shape(d)[0])
            win2 = win(np.shape(d)[1])
            win = np.outer(win1,win2)
            win = win/np.sqrt(np.mean(win**2))

    #Compute Fourier components
    return np.fft.fftn(d*win)/np.size(d)
    
def realPSD(d0,win=np.hanning,dx=1.,axis=None,nans=False,minpx=10):
    """(Ryan Allured) This function returns the PSD of a real function
    Gets rid of zero frequency and puts all power in positive frequencies
    Returns only positive frequencies
    """
    from utilities.imaging.man import stripnans
    if nans is True:
        d = stripnans(d0)
    else:
        d = d0
    if len(d) < minpx:
        return np.nan
    #Get Fourier components
    c = components(d,win=win)
    #Handle collapsing to 1D PSD if axis keyword is set
    if axis==0:
        c = c[:,0]
    elif axis==1:
        c = c[0,:]

    #Reform into PSD
    if np.size(np.shape(c)) == 2:
        f = [np.fft.fftfreq(np.shape(c)[0],d=dx)[:np.shape(c)[0]/2],\
                   np.fft.fftfreq(np.shape(c)[1],d=dx)[:np.shape(c)[1]/2]]
        c = c[:np.shape(c)[0]/2,:np.shape(c)[1]/2]
        c[0,0] = 0.
        #Handle normalization
        c = 2*c
        c[0,:] = c[0,:]/np.sqrt(2.)
        c[:,0] = c[:,0]/np.sqrt(2.)
        
    elif np.size(np.shape(c)) == 1:
        f = np.fft.fftfreq(np.size(c),d=dx)
        f = f[:np.size(c)/2]
        c = c[:np.size(c)/2]
        c[0] = 0.
        c = c*np.sqrt(2.)

    return f[1:],np.abs(c[1:])**2

   
def xrealPSD(d,win=1,dx=1.,axis=None):
    """RA This function returns the PSD of a real function
    Gets rid of zero frequency and puts all power in positive frequencies
    Returns only positive frequencies
    """
    #Get Fourier components
    c = components(d,win=win)  #np.fft.fftn(d*win)/np.size(d)
    #Handle collapsing to 1D PSD if axis keyword is set
    if axis==0:
        c = c[:,0]
    elif axis==1:
        c = c[0,:]

    #Reform into PSD
    if np.size(np.shape(c)) is 2:
        f = [np.fft.fftfreq(np.shape(c)[0],d=dx)[:np.shape(c)[0]/2],\
                   np.fft.fftfreq(np.shape(c)[1],d=dx)[:np.shape(c)[1]/2]]
        c = c[:np.shape(c)[0]/2,:np.shape(c)[1]/2]
        c[0,0] = 0.
        #Handle normalization
        c = 2*c
        c[0,:] = c[0,:]/np.sqrt(2.)
        c[:,0] = c[:,0]/np.sqrt(2.)
        
    elif np.size(np.shape(c)) is 1:
        f = np.fft.fftfreq(np.size(c),d=dx)
        f = f[:np.size(c)/2]
        c = c[:np.size(c)/2]
        c[0] = 0.
        c = c*np.sqrt(2.)

    return f,np.abs(c)**2    
import pdb

def plot_psd(f,p,units=None,label=None,span=0,psdrange=None,
    includezerofreq=False,*args,**kwargs):
    """Plot a PSD on logaritmic axis with standard labels.
    Units can be passed as profile units (2-el) or x,y,z (3-el, only last two are used). 
    If span is set, plots psd
    range according to span settings of data2D.projection."""
    #Note: title is not implemented because it can be a suptitle, so it is left to set
    # to caller routine. Take care to manage multiple plotting as list or successive calls.
    #this gets very tricky to do
    

    if label is not None:
        l = label
    else:
        l = ""
    
    if not includezerofreq:
        if f[0]==0:
            f=f[1:]
            p=p[1:,...]
            
    if units is None: units=['[X]','[Y]','[Z]']
    if len(units)==2: units = [None,units[0],units[1]]
    #print ("FIX THIS ROUTINE BEFORE USING IT, see newview_plotter")
    plt.ylabel('axial PSD ('+units[2]+'$^2$ '+units[1]+')')
    plt.xlabel('Freq. ('+units[1]+'$^{-1}$)')
    #pdb.set_trace()
    if len(p.shape)==2: #experimentally deals with 2D array automatically plotting average
        ps=projection(p,axis=1,span=span)
        if span:
            plt.plot(f,p[0],label=l + 'AVG')
            plt.plot(f,p[1],label=l + 'point-wise min')
            plt.plot(f,p[2],label=l + 'point-wise max')
        else:
            plt.plot(f,p,label=l + 'AVG')
    else:
        assert len(p.shape)==1
        plt.plot(f,p,label=l,*args,**kwargs)
    
    if psdrange is not None:
        if len(psdrange)==1: psdrange=np.repeat(psdrange,2)
        plt.ylim(psdrange)
    plt.loglog()
    plt.grid(1)
    plt.legend(loc=0) 
    #plt.suptitle('%s: rms=%5.3f 10$^{-3}$'%(label,rms*1000)+units[2])

from matplotlib import pyplot as plt
import numpy as np
from cycler import cycler
from collections import OrderedDict

def make_psd_plots(toplot,units=None,outfile=None): 

    """This was in makenicePSDplots.make_plots, needs to be integrated with psd_plot.
    
    toplot is a dictionary (or collections.OrderedDict if order is important) with filenames as keys and labels as values.
    psd is read from each file on the first two columns and plotted."""
    
    plt.figure()
    # 1. Setting prop cycle on default rc parameter
    plt.rc('lines', linewidth=1.5)
    plt.rc('axes', prop_cycle=(cycler('linestyle', ['-', '--', ':', '-.'])* cycler('color', ['r', 'g', 'b', 'y']) ))
       
    for ff,ll in toplot.items():    
        f,p=np.genfromtxt(ff,delimiter='',unpack=True,usecols=[0,1])
        plot_psd(f,p,label=ll,units=units)
        """
        plt.plot(f,p,label=ll)
        plt.loglog()
        plt.xlabel('freq. (mm$^-1$)')
        plt.ylabel('mm $\mu$m$^2$')
        plt.grid(1)
        plt.legend(loc=0)
        plt.show()
        """
        
    if outfile is not None:
        plt.savefig(outfile)
        print("results saved in %s"%outfile)
        #plt.savefig(fn_add_subfix(ff,'_psdcomp','.png'))
    import matplotlib
    matplotlib.rcParams.update(matplotlib.rcParamsDefault)
    
def linearTrend(x,y): 
    """test linear trend"""
    yp=y-np.mean(y)
    yt=y-line(x,y)
    
    plt.clf()
    for yy,ll in zip([y,yp,yt],['signal','avg offset','line removed']):
        plot_sig_psd(x,yy,label=ll)
    for ax in plt.gcf().axes:
        plt.sca(ax)
        plt.legend(loc=0)

class PSDplot(object):
    """Contains a number of psds in form (f,psd)."""
    def __init__(self):
        self.psds=[]
        #self.labels=[]
        self.kwargs=[]
        
    def append(self,psd):
        self.psds.append(psd)
        
    def append_from_file(self,file,cols,**kwargs):
        self.psds.append(np.genfromtxt(file,skip_header=1,unpack=1)[[cols[0],cols[1]]])       
        self.kwargs.append(kwargs)
        #self.labels.append(label)
        
        
    def plot(self,*args):
        plt.figure()
        for i,(yfreq1,ypsd1) in enumerate(self.psds):
            #label=kwargs.pop('label',None)
            plt.plot(yfreq1,ypsd1,*args,**(self.kwargs[i]))
        plt.loglog()
        plt.grid(1)
        plt.legend(loc=0)
        

def addPSDs(psdfile,cols,*args,**kwargs):
    """Add to current plot a list of PSDs from a standard (12-col,1 line header) PSD file.
    """
    
    assert len(cols)==2
    
    for psdfile1,cols1 in zip(psdfiles,cols):
        yfreq1,ypsd1=np.genfromtxt(psdfile1,skip_header=1,unpack=1)[cols[0],cols[1]]
    
    #plt.figure()
    plt.plot(yfreq1,ypsd1,*args,**kwargs)
    #plt.loglog()
    #plt.grid()
    #if title is None:
    #    title=os.path.basename(psdfile1)
    #plt.title(title)
    plt.legend(loc=0)
    if outfolder:
        plt.savefig(os.path.join(outfolder,title+'.png'))    

def plotPSDs(psdfile1,psdfile2,title=None,outfolder=None):
    """Plot a PSD from a standard (12-col,1 line header) PSD file."""
    
    yfreq1,ypsd1,xfreq1,xpsd1=np.genfromtxt(psdfile1,skip_header=1,unpack=1)[8:]
    yfreq2,ypsd2,xfreq2,xpsd2=np.genfromtxt(psdfile2,skip_header=1,unpack=1)[8:]
    
    plt.figure()
    plt.plot(yfreq1,ypsd1,'b',label='before annealing')
    plt.plot(xfreq1,xpsd1,'b')
    plt.plot(yfreq2,ypsd2,'r',label='after annealing')
    plt.plot(xfreq2,xpsd2,'r')
    plt.loglog()
    plt.grid()
    plt.ylabel('(mm $\mu$m^2)')
    plt.xlabel('mm^-1')
    if title is None:
        title=os.path.basename(psdfile2)
    plt.title(title)
    plt.legend(loc=0)
    if outfolder:
        plt.savefig(os.path.join(outfolder,title+'.png'))           


def FFTtransform(x,y):
    yfft  = np.fft.rfft(y) #note that rfft return exact values as fft (needs to be doubled after squaring amplitude)
    N = len(y)
    L=span(x,True)  
    f = np.fft.rfftfreq(N,L/N)
    return f,yfft

def plot_transform(f,yfft,label="",euler=False,units=None,includezerofreq=False,**kwargs):
    """units is a 2 element vector."""
    #pdb.set_trace()
    ax3 = plt.gca()
    ##PLOT TRANSFORM AS EULER REPRESENTATION  
    unitstr = psd_units(units) 
    
    if not includezerofreq:
        if f[0]==0:
            f=f[1:]
            yfft=yfft[1:,...]
            
    if euler:
        plt.title('DFT - Module and phase')
        plt.plot(f,np.abs(yfft),label=label,**kwargs)
        #plt.loglog() 
        ax3.legend(loc=2)
        #string for x is set to '' as default:
        plt.xlabel('Freq ('+(unitstr[0] if unitstr[0] else '[X]')+'$^-1$)')
        plt.ylabel('Module'+((' ('+unitstr[1]) if unitstr[1] else '[Y]')+'$^2$)')
            
        ax3b = ax3.twinx()
        ax3.twin=ax3b
        ax3b.plot(f,np.angle(yfft)/np.pi,'--',label=label+' Phase',
        **kwargs)
        ax3b.set_ylabel('Phase /$\pi$')
        ax3b.set_ylim(-1.,1.)
        plt.grid(1)
        ax3b.legend(loc=1)
        
    ##PLOT TRANSFORM AS COMPLEX NUMBER
    else:
        #pdb.set_trace()
        plt.title('DFT - Real and imaginary parts')
        ax3.plot(f,np.real(yfft),label=label+' Re',**kwargs)
        #plt.loglog() 
        s = span (np.real(yfft))
        plt.ylim([s[0],s[1]]) # needed to expland scale if very small
        #ax3.set_xlabel(unitstr[0]+'$^-1$')
        #string for x is set to '' as default:
        plt.xlabel('Freq. ('+(unitstr[0] if unitstr[0] else '[X]')+'$^-1$)')
        ax3.set_ylabel('Re')
        #for tl in ax2.get_yticklabels():
            #tl.set_color('b')
        ax3.legend(loc=2)
            
        ax2b = ax3.twinx()
        ax2b.plot(f,np.imag(yfft),'--',label=label+' Im',**kwargs)
        s = span (np.imag(yfft))
        plt.ylim([s[0],s[1]]) # needed to expland scale if very small
        ax2b.set_ylabel('Im')
        #for tl in ax2b.get_yticklabels():
            #tl.set_color('r')
        plt.grid(True)
        ax2b.legend(loc=1)
        ax3.twin=ax2b
    plt.tight_layout

def xplot_sig_psd(x,y,label="",**kwargs):
    """plot signal, fft and psd on three panels.
    dft sets how the fft is plotted 
    'realim': real and im parts,
    'euler' argument and phase
    'both' uses two panels (four in total).
    """ 
    N = len(y)
    L=span(x,True)    
        
    ax1=plt.subplot(211)
    plt.title('Signal')
    plt.plot(x,y,**kwargs)
    ax1.set_xlabel('mm')
    plt.grid(1)
    
    ax4=plt.subplot(212) 
    plt.title('PSD')
    f,PSD=psd(x,y)
    plt.plot(f,PSD,label=label,**kwargs)
    plt.loglog()     
    ax4.set_xlabel('mm^-1')
    plt.grid(1)
    plt.show()
        
    return ax1,ax4
    
def plot_sig_psd(x,y,label="",realim=False,
    euler=False,power=True,includezerofreq=False,
    yrange=None,prange=None,fignum=None,
    aspect='auto', units=None,outname=None,norm=1,rmsnorm=True,**kwargs):
    """plot signal, fft and psd on n panels according to flags, return axis.
    for axis with a twin axis, the twin axis is appended to the first in a property 'twin'
    parameter for plot function (styles etc., common to all plots) 
        can be passed as kwargs
    plots signal and  additional plots according to respective flags:
    'realim': real and im parts,
    'euler' argument and phase
    'power' plot PSD calling psd with default parameters
    """ 
    #can be done in more advanced way with ax.change_geometry(numrows, numcols, num)
    nsp=1+np.count_nonzero([realim,euler,psd]) #number of subplots
    i=1 #index of subplot
    axes=[]
    
    #ax1 SIGNAL
    axes.append(plt.subplot(nsp,1,i))
    i=i+1
    plt.title('Signal') 
    plt.plot(x,y,**kwargs)
    plt.xlabel('mm')
    plt.grid(1)
    
    f,yfft =  FFTtransform(x,y)
    
    '''
    ##calculate
    N = len(y)
    L=span(x,True)    
    #yfft  = np.fft.fft(y)
    #yfft  = yfft[:int(np.ceil(N/2))] #select positive part of spectrum
    yfft  = np.fft.rfft(y) #note that rfft return exact values as fft (needs to be doubled after squaring amplitude)
    #f = np.fft.fftfreq(N,L/N)
    #f = f[:int(np.ceil(N/2))]
    f = np.fft.rfftfreq(N,L/N)
    '''
    
    '''
    The default normalization has the direct transforms unscaled and the inverse transforms are scaled by 1/n. It is possible to obtain unitary transforms by setting the keyword argument norm to "ortho" (default is None) so that both direct and inverse transforms will be scaled by 1/sqrt{n}.
    
    N = len(y)
    L=span(x,True)
    #form = 1 --> normfactor = 1./N**2*L
    
    yfft  = np.fft.rfft(y*win) #note that rfft return exact values as fft (needs to be doubled after squaring amplitude)
    
    normfactor=normPSD(N,L,form=norm)
    
    if rmsnorm:  #normalize to rms of non windowed function. This is independent on 0 constant term, 
    # but also alter its meaning 
        normfactor=normfactor*np.nanstd(y)**2/np.nanstd(y*win)**2
    
    psd  = 2*normfactor*np.abs(yfft)**2
    psd[0]=psd[0]/2  #the first component is not doubled
    
    freqs=np.fft.rfftfreq(N,np.float(L)/(N-1))
    '''
        
    if realim:
        ax2=plt.subplot(nsp,1,i)
        axes.append(ax2)
        i=i+1
        plot_transform(f,yfft,label=label,includezerofreq=includezerofreq,**kwargs)
        #for tl in ax2.get_yticklabels():
            #tl.set_color('b')
        
    if euler:
        ax3=plt.subplot(nsp,1,i)  
        axes.append(ax3)
        i=i+1        
        plot_transform(f,yfft,label=label,includezerofreq=includezerofreq,euler=True,**kwargs)
    
    ##PLOT PSD
    if power:
        ax = plt.subplot(nsp,1,i)
        axes.append(ax)
        i=i+1
        plt.title('PSD')
        f,PSD=psd(x,y,**kwargs)
        plt.plot(f,PSD,label=label,**kwargs)
        plt.loglog()     
        plt.xlabel('mm^-1')
        plt.grid(1)
        plt.ylim(prange)
    
    plt.tight_layout()
    #plt.show()
    return axes
    
def plot_sig_psd4(x,y,scale=1.0,label="",**kwargs):
    """ wrapper around`plot_sig_psd4`
    plot signal, fft and psd on three panels.
    like `plot_sig_psd`, with fixed number of variables.
    
    ?? 
    dft sets how the fft is plotted 
    'realim': real and im parts,
    'euler' argument and phase
    'both' uses two panels (four in total).
    """ 
    
    print ("WARNING: `plot_sig_psd4` was replaced with `plot_sig_psd`, update code!")
    return plot_sig_psd(x,y,label=label,realim=True,
    power=True,**kwargs)
'''        
def plot_sig_psd4(x,y,scale=1.0,label="",**kwargs):
    #replaced by the new plot_sig_psd
    """plot signal, fft and psd on three panels.
    like `plot_sig_psd`, with fixed number of variables.
    
    ?? 
    dft sets how the fft is plotted 
    'realim': real and im parts,
    'euler' argument and phase
    'both' uses two panels (four in total).
    """ 
    N = len(y)
    L=span(x,True)    
        
    ax1=plt.subplot(411)
    plt.title('Signal')
    plt.plot(x,y,**kwargs)
    ax1.set_xlabel('mm')
    plt.grid(1)

    ax2=plt.subplot(412) 
    yfft  = np.fft.fft(y)
    yfft  = yfft[:np.ceil(N/2)]
    f = np.fft.fftfreq(N,L/N)
    f = f[:np.ceil(N/2)]
    
    plt.title('DFT - Real and imaginary parts')
    ax2.set_ylabel('time (s)')
    ax2.plot(f,np.real(yfft),label=label+' Re',**kwargs)
    #plt.loglog() 
    ax2.set_xlabel('mm^-1')
    ax2.set_ylabel('Re')
    #for tl in ax2.get_yticklabels():
        #tl.set_color('b')
    plt.legend(loc=0)
		
    ax2b = ax2.twinx()
    ax2b.plot(f,np.imag(yfft),'--',label=label+' Im',**kwargs)
    ax2b.set_ylabel('Im')
	#for tl in ax2b.get_yticklabels():
		#tl.set_color('r')
    plt.grid(True)
    plt.legend(loc=0)
    
    ax3=plt.subplot(413)
    plt.title('DFT - Module and phase')
    ax3.set_ylabel('time (s)')
    ax3.plot(f,np.abs(yfft),label=label+' Norm',**kwargs)
    #plt.loglog() 
    plt.legend(loc=0)
    ax3.set_xlabel('mm^-1')
    ax3.set_ylabel('Norm')
    	
    ax3b = ax3.twinx()
    ax3b.plot(f,np.angle(yfft)/np.pi,'--',label=label+' Phase',
    **kwargs)
    ax3b.set_ylabel('Phase /$\pi$')
    ax3b.set_ylim(-1.,1.)
    plt.grid(1)
    plt.legend(loc=0)
    plt.show()
    
    ax4=plt.subplot(414) 
    plt.title('PSD')
    f,PSD=psd(x,y,scale=scale)
    plt.plot(f,PSD,label=label,**kwargs)
    plt.loglog()     
    ax4.set_xlabel('mm^-1')
    plt.grid(1)
    plt.show()
        
    return ax1,ax2,ax3,ax4
'''
  
def testPSDest(x,y): 
    """plot 3 panels with: test profile, PSD calculated with different functions 
    and PSD calculated by piecewise matplotlib function, with different settings.
    return 3 axis."""
    # 2016/04/04 corrected for updated normalization
    
    
    plt.clf()
    plt.suptitle('Comparison of PSD estimations')
    ax1=plt.subplot(311)
    plt.title('Signal=wave+line+noise')
    plt.plot(x,y)

    ax2=plt.subplot(312)
    plt.title('VC and RA routines')
    f,PSD=psd(x,y-line(x,y))
	#plt.plot(f,PSD/N/L,label='my PSD/N/L (line subtraction)')
    plt.plot(f,PSD,label='my PSD (line subtraction)')
    fw,PSDw=psd(x,y*np.hanning(len(y)))
	#plt.plot(fw,PSDw/N/L,label='my PSD windowed/N/L')
    plt.plot(fw,PSDw,label='my PSD windowed')
    fw,PSDw=psd(x,(y-line(x,y))*np.hanning(len(y)))
	#plt.plot(fw,PSDw/N/L,label='my PSD windowed/N/L (line subtraction)')
    plt.plot(fw,PSDw,label='my PSD windowed (line subtraction)')

    plt.plot(*realPSD(y*np.sqrt(N*L),win=1,dx=L/N),label='realPSD/df (RA)',linestyle='-.',color='r')    
    plt.plot(*realPSD(y*np.sqrt(N*L),dx=L/N),label='realPSD/df window (RA)',linestyle='-.',color='k')

    #plt.plot(*realPSD(y,dx=L/(N-1)),label='realPSD',linestyle='-.',color='r')    
    #plt.plot(*realPSD(y*np.hanning(len(y)),dx=L/(N-1)),label='realPSD - w',linestyle='-.',color='k')
    plt.loglog()
    plt.legend(loc=0)
    
    ax3=plt.subplot(313)
    from matplotlib.pyplot import psd as psdplot
    from matplotlib.mlab import window_none
    plt.title('Matplotlib piecewise estimation')
    detrending=['default' ,'constant' , 'mean' , 'linear' , 'none']
    for det in detrending:
        pp,ff=psdplot(y,Fs=N/float(L),detrend=det,window=window_none,scale_by_freq=True)
    plt.legend(detrending,loc=0)
    plt.loglog()
    plt.show()
    return ax1,ax2,ax3


def test_psd_normalization(x,y,wfun=None,norm=1,**kwargs):
    """Calculate PSD with a given normalization and
    compare its integral over frequency and sum to
    rms square and standard deviation (they differ for rms including also offset, while stddev being referred to mean."""
    
    f,p=psd(x,y,wfun=wfun,norm=norm,**kwargs)
    
    print ("== Quantities calculated from profile ==")
    print ("Profile Height PV %6.3g (min: %6.3g, max: %6.3g)"%
           (np.nanmax(y)-np.nanmin(y),np.nanmin(y),np.nanmax(y)))
    print ("Profile Height avg.: ",np.nanmean(y))
    print ("devstd**2=",np.std(y)**2, '(devstd=%f5.3)'%np.std(y))
    print ("rms**2=",(y**2/len(x)).sum())  #, '(rms=%f5.3)'%np.std(y)
    print ("\n== Quantities calculated from PSD ==")
    print ("sum of PSD is ",p[1:].sum())
    print ("integral of PSD (as sum*deltaf) is ",p[1:].sum()/span(x,1)) #span(x,size=1)=1/L
    print ("integral trapz: ",np.trapz(p[1:],f[1:]))
    print ("psd[0]*deltaf=%f (integral including:%f)"%(p[0]*f[1],p.sum()/span(x,1)))
    print ("#--------\n")
    
    return f,p          
    
        
if __name__=="__main__":
    N=500
    noise=2.
    amp=10.
    L=30.
    nwaves=5.8 #32
    ystartend=(3,20)
    #ystartend=(0,0)
    
    x,y=make_signal(amp,L=L,N=N,nwaves=nwaves,ystartend=ystarted,noise=noise)
    ax1,ax2,ax3=testPSDest(x,y)   
    
    plt.sca(ax2)
    plt.axvline(nwaves/L)    
    plt.sca(ax3)
    plt.axvline(nwaves/L)  
    plt.draw()
    for nn in range(1,3):
        print("norm type= %s"%nn)
        test_psd_normalization(x,y,norm=nn)
 



"""
2020/09/21
Differently from what done in pySurf, here we keep reader and register mechanism separated and leave to the user the responsibility to call them. They can be put together in future if a convenient model for readers is found.

`pyProfile.profile` has only a simple reader (a thin wrapper around `np.genfromtxt`) and a function `register_profile` to align and rescale x and y. Additional readers (`read_mx_profiles`, `read_xyz`) are considered experimental and kept here, even if they are functions.

2020/06/25 start to create profile class on template of data2D_class.
N.B.: there is an important difference between profile and data2D functions
work, and is that data2D functions accept as argument surfaces in the form (tuple)
`(data,x,y)`, while functions in `profile` want x and y as separate arguments.
Meaning that in this second case, passing only one argument is check with None
as default val on second argument, rather than with inspecting the elements together.
"""


"""
notes from data2D_class.py:
2018/06/07 v1.3
v1.2 was not convenient, switch back to same interface as 1.1:
don't modify self, return copy.

2018/06/06 v 1.2
After attempt to modify interface in a way that modify self, switch back to same interface as 1.1. It is inconvenient to have always to create a copy, in particular in interactive mode when same operation can be repeated multiple times you need to inizialize data every time (e.g. rotate).

2018/06/06 v1.1
methods are written to consistently return a copy without modifying self.
But this doesn't work, always need to assign, cannot chain (with property assignment, it works on methods) or apply to set of data in list.

Give some attention to inplace operators that can link to external data. At the moment when class is created from data, x, y these are assigned directly to the property resulting in a link to the original data. Some methods have inplace operations that will reflect on initial data, others don't."""

"""
programming notes:
-inplace methods-
A method acting on Data2D (e.g. level) can modify (o reassign to) self and return self or return a copy.

e.g. a method crop (or remove nancols) that return slices of data and x/y reassign to self.data/x/y views of original data.

A method that reassigns self.data=...
most of the time stays linked to original data.
If inplace changes (e.g. element-wise assignment) are then performed on the property, the change is reflected in the original data.
See notebook Python Programming Notes.

To remove this behavior, a copy of the data or view must be performed at some point (e.g. explicitly self.data.copy() or implicitly deepcopy(self).



-attempt to subclass ndarray-
def __new__(subtype, shape, dtype=float, buffer=None, offset=0,
            strides=None, order=None, info=None):
    # Create the ndarray instance of our type, given the usual
    # ndarray input arguments.  This will call the standard
    # ndarray constructor, but return an object of our type.
    # It also triggers a call to InfoArray.__array_finalize__
    obj = super(Data2D, subtype).__new__(subtype, shape, dtype,    #(InfoArray
                                            buffer, offset, strides,
                                            order)
    # set the new 'info' attribute to the value passed
    #obj.info = info
    # Finally, we must return the newly created object:
    return obj

def __array_finalize__(self, obj):
    # ``self`` is a new object resulting from
    # ndarray.__new__(InfoArray, ...), therefore it only has
    # attributes that the ndarray.__new__ constructor gave it -
    # i.e. those of a standard ndarray.
    #
    # We could have got to the ndarray.__new__ call in 3 ways:
    # From an explicit constructor - e.g. InfoArray():
    #    obj is None
    #    (we're in the middle of the InfoArray.__new__
    #    constructor, and self.info will be set when we return to
    #    InfoArray.__new__)
    if obj is None: return
    # From view casting - e.g arr.view(InfoArray):
    #    obj is arr
    #    (type(obj) can be InfoArray)
    # From new-from-template - e.g infoarr[:3]
    #    type(obj) is InfoArray
    #
    # Note that it is here, rather than in the __new__ method,
    # that we set the default value for 'info', because this
    # method sees all creation of default objects - with the
    # InfoArray.__new__ constructor, but also with
    # arr.view(InfoArray).
    #self.info = getattr(obj, 'info', None)
    # We do not need to return anything
"""



import os
from pathlib import Path
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy
import pdb
import inspect  # to use with importing docstring

from dataIO.span import span
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.superlist import Superlist
from dataIO.arrays import split_blocks, split_on_indices
from dataIO.functions import update_docstring

from pySurf.affine2D import find_affine
from pyProfile import profile
from pyProfile.profile import crop_profile
from pyProfile.profile import level_profile
from pyProfile.profile import resample_profile
from pyProfile.profile import sort_profile
from pyProfile.profile import sum_profiles, subtract_profiles
from pyProfile.psd import psd as profpsd
from pyProfile.profile import movingaverage, rebin_profile
from pyProfile.profile import merge_profiles
from pyProfile.profile import register_profile
from pyProfile.profile import save_profile


'''
def update_docstring(func,source):
    """given a current function and a source function, update current function docstring
     appending signature and docstring of source.

     It provides a decorator @update_docstring(source) update the decorated
       function docstring. User should take care to create docstring of decorated function
       only as preamble to source docstrings, e.g.:
       this is func function, derived from source function by modifying the usage of parameter foo.
       parameter foo is changed of sign."""
    doc0="" if func.__doc__ is None else func.__doc__
    func.__doc__='\n'.join([doc0,source.__name__+str(inspect.signature(source)),source.__doc__])
    return func

def doc_from(source):
    """intent is to use partial to obtain a decorator from update_docstring.
    not sure it is the right way."""
    partial(update_docstring,func=func)
'''

def read_xyz(filename,*args,**kwargs):
    """temptative routine to read xyz files.
    Use by reading the data and then passing them to profile.
    It will be incorporated in some form of reader in a 
    more mature version"""

    raise NotImplementedError
    
def read_mx_profiles(filename,*args,**kwargs):
    """temptative routine to read xyz files. Return a list of all profiles
        with names from 3rd column (files are in format x,y,profilename).
    Use by reading the data and then passing them to profile.
    It will be incorporated in some form of reader in a 
    more mature version"""

    x,y,tag = np.genfromtxt(filename,delimiter=',',
                            unpack=True,skip_header=1,dtype=str,*args,**kwargs)
    x=x.astype(np.float)
    y=y.astype(np.float)
    labels = np.unique(tag)  # names of profiles in order
    groups = [np.where(tag==l)[0] for l in labels] # indices of points for each profile 
    profiles = [Profile(x[i],y[i],name=tag[i[0]]) for i in groups]
    
    return profiles


def get_header_field(db, tag, header='<<PMCA SPECTRUM>>'):
    """Extract a single field named `tag` as a list.
    tag must should match a field (column) name as inferred from header 
    (otherwise empty list is returned).
    Temporarily here for use with `read_mca`, can be moved to a more general
    use in dataIO.
    """
    
    from dataIO.read_pars_from_namelist import namelist_from_string
    headers = [d.header[header] for d in db['data']]
    return [namelist_from_string(t,separator='-')[tag] for t in headers]
    
def read_mca(filename,*args,**kwargs):
    """ Given a .mca file, return a `Profile` object.
    
    temptative routine to read mca files from amptek energy sensitive detector. Return a profile with metadata information in a `.header` property (temporarily a dictionary obtained from string blocks) of all profiles.
    Like all other readers will be incorporated in some form of reader in a 
    more mature version.
    """

    import re
    from scipy import interpolate

    a = open(filename,'r').readlines()
    a = [aa.strip() for aa in a if len(aa.strip())]

    p=re.compile("<<.*>>")
    i = p.match("".join(a))

    itags = [i for i,l in enumerate(a) if re.compile("<<.*>>").match(l)] #posizione dei tags in linee
    tags = [a[i] for i in itags]  #tags
    
    blocks = {'<<CALIBRATION>>':['LABEL - Channel','0 0.','1 1']} #default calibration if not defined in file, in a consistent format for conversion.
    for i,t in enumerate(tags[:-1]):  #last tag is assumed to be closing tag
        if itags[i+1] != itags[i]+1:
            blocks [t] = a[itags[i]+1:itags[i+1]]  
    #pdb.set_trace()
    #print(blocks['<<CALIBRATION>>'])
    data = [float(d) for d in blocks['<<DATA>>']  ]
    cal = np.array([[float(dd) for dd in d.split()] for d in blocks['<<CALIBRATION>>'][1:]])
    #print(len(data),cal)
    x = interpolate.interp1d(cal[:,0],cal[:,1],fill_value='extrapolate')(np.arange(len(data))) #use interp1d because np.interp doesn't extrapolate
    #print(span(x),len(x))
    
    #print(cal)
    profiles = Profile(x,data,name=filename)
    profiles.header = blocks
    
    return profiles

    
class Profile(object):  #np.ndarrays
    """A class containing x,y data. It has a set of methods for analysis and visualization.
    Function methods: return a copy with new values.
    to use as modifier (procedure), e.g.:
        a=a.level() """

    """C'era qui un tentativo fallito di subclassare nd.array
    usando __new__ e __array_finalize__."""


    def __init__(self,x=None,y=None,file=None,reader=None,
        scale = (1.,1.), units=None,name=None,*args,**kwargs):
        """can be initialized with data; x,y; file; file, x
        if x is provided, they override x from data if matching number of elements, 
           or used as range if two element (error is raised in case of ambiguity)."""
        
        #from pySurf.readers.instrumentReader import reader_dic
        from pyProfile.profile import register_profile
        
        #pdb.set_trace()

        if isinstance (x,str):
            print ('first argument is string, use it as filename')
            file=x
            x=None
        else:
            if y is not None:
                y=np.array(y) #onvert to array if not
            
        #pdb.set_trace()
        self.file=file #initialized to None if not provided
        if file is not None:
            assert y is None
            # passed file AND xrange, overrides x because only range matters.
            #store in xrange values for x if were passed
            xrange=span(x) if x is not None else None
            #pdb.set_trace()
            
            if reader is not None: raise NotImplementedError("readers are not implemented yet for profiles,"+
                "\tPass data or read from two column text file in compatible format.")
            
            #pdb.set_trace()
            self.load(file,*args,**kwargs)
            """
            if reader is None:
                reader=auto_reader(file) #returns a reader
            """
            from pyProfile.profile import load_profile
            x,y=load_profile(file,*args,**kwargs) #calling without arguments skips register, however skips also reader argumnets, temporarily arranged with pop in read_data to strip all arguments for
                #register data and pass the rest to reader
            #pdb.set_trace()
            #TODO: Cleanup this part and compare with pySurf. Consider centering
            # of axis coordinates as in grid example.
            if np.size(xrange) == np.size(y):
                # x is compatible with data size
                if np.size(xrange) == 2:
                    print ('WARNING: You passed both data and x, with size 2. There may be some ambiguity in how coordinates are handled')
                self.x=xrange
            elif np.size(xrange) == 2:
                #here it must be intended as a range. Should check also interval centerings.
                #it was considered somewhere in pySurf.
                x=np.linspace(*xrange,np.size(y))
            elif xrange is not None:
                print("wrong number of elements for x (must be 2 or xsize_data), it is instead",np.size(xrange))
                raise ValueError
            
            # set self.header to file header if implemented in reader, otherwise set to empty string""
            #questo dovrebbe sempre fallire
            try:
                #kwargs['header']=True
                self.header=reader(file,header=True,*args,**kwargs)
            except TypeError:  #unexpected keyword if header is not implemented
                self.header=""
                #raise
        else:
            if y is not None:
                if len(y.shape) != 1:
                    #pdb.set_trace()
                    print('WARNING: data are not uniidimensional, results can be unpredictable!')
                if x is None:
                    x=np.arange(np.size(y))

    #if data is not None:
        
        x,y=register_profile(x,y,scale=scale,*args,**kwargs)# se load_profile calls register, this
        #goes indented.

        self.x,self.y=x,y

        if np.size(units) == 1:
            units=[units,units]
        self.units=units
        if name is not None:
            self.name=name
        elif file is not None:
            self.name=os.path.basename(file)
        else:
            self.name=""
        #print(name)

    def __call__(self):
        return self.x,self.y
    """
    def __add__(self,other,*args,**kwargs):
        return Profile(*sum_profiles(*self(),*other(),*args,**kwargs),units=self.units,name=self.name + " + " + other.name)
    """

    def __add__(self,other,*args,**kwargs):
        
        if isinstance(other,Profile):
            res = sum_profiles(*self(),*other(),*args,**kwargs)
            res = Profile(*res,units=self.units,name=self.name + " + " + other.name)
        else:
            try:
                res = self.copy()
                res.y = res.y + other
                res.units = self.units
            except ValueError:
                raise ValueError("Unrecognized type in sum")
        return res
        
    def __sub__(self,other,*args,**kwargs):
        """
        This is made on copy and paster from add. It is maybe not the most flexible way in relation with multiplication and operations with different types, but it works well enough.
        
        original code (handle only profile objects):       
            res=Profile(*(subtract_profile(*self(),*other(),*args,**kwargs))(),units=self.units)
            res.name = self.name + " - " + other.name
            return res
        """
        #  y = np.genfromtxt(fn)
        #p=Profile(np.arange(len(y)),y)
        #p.plot(label = 'raw')
        #
        #p.level().plot('r',label = 'leveled')
        #(p-p.level()).plot('r',label = 'fit')
        #plt.legend()
        #
        #pdb> isinstance(other,Profile)
        #   False

        #ipdb> other
        #   <pyProfile.profile_class.Profile object at 0x000002A55C104CF8>

        #ipdb> Profile
        #<class 'pyProfile.profile_class.Profile'>
        """
        print ("class: ",other.__class__)
        # Out[65]: pyProfile.profile_class.Profile

        print("check: ",other.__class__ == Profile)
        # Out[66]: True

        print ("type: ",type(other))
        # Out[65]: pyProfile.profile_class.Profile

        print("check: ",type(other) == Profile)
        #Out[67]: True

        print("check: ", isinstance(other,Profile))
        #Out[68]: True
        """
        
        #pdb.set_trace()
        
        if isinstance(other,Profile):
            res = subtract_profiles(*self(),*other(),*args,**kwargs)
            res = Profile(*res,units=self.units,name=self.name + " - " + other.name)
        else:
            try:
                res = self.copy()
                res.y = res.y - other
                res.units = self.units
            except ValueError:
                raise ValueError("Unrecognized type in subtraction")
        return res
    
    def __repr__(self):
    
        return '<.Profile "%s" at %s>'%(self.name,hex( id(self)))
        
        ''' N.B.: you cannot retrieve and modify existing __repr__,
        the following gives infinite recursion, the only way is to recreate a new repr: 
       
            class me():
                def __repr__(self):
                    return self.__repr__()
                    
            a = me()
            a'''
    
    
    def __mul__(self,scale,*args,**kwargs):
        """mutiplocation: accept Profile, scalars or 2-vector [x,y]. """
        
        # this is called directly if self is a Profile object.
        # mul gets called if self and scale are same type,
        # or if self can handle self.__mul__(scale)
        # when this doesn't work, it calls scale.__rmul__(self)
        # so for example:
        # p1*p2 calls p1.mul(p2) -> ok
        # p2*5 calls p2.mul(5) -> ok
        # 5*p2 calls p2.__rmul__(5) <-- fallback from 5.__mul__(p2)  
        
        res = self.copy()
        #breakpoint()
        if np.size(scale)==1:  # scalar object or value
            if isinstance(scale,Profile):
                """if it is Profile, do pointwise multiplication rescaling on firts."""
                # resample and multiply. For surface result, use matrix multiplication.
                #raise NotImplementedError ('can be ambigous (return point to point multipl. or surface? Fix in code, at the momeb accept only x and (y) scalars.')
                tmp=scale.resample(self)
                res=self.copy()
                res.y = self.y * tmp.y
                
                #FIXME questo non convince, ad es, se none, ad es se divisione
                if scale.units[1] == self.units[1]:
                    res.units[1] = scale.units[1]+'^2'
                else:
                    res.units[1]='%s %s'%(self.units[1], scale.units[1])
                if self.name and scale.name:
                    res.name = self.name + ' x ' + scale.name
                else: 
                    res.name = (self.name if self.name else scale.name) + ' product'
            else:     # value
                breakpoint()
                res.y = scale * res.y 
                if self.name:
                    res.name = '%s x %s'%(self.name,scale)
                else:
                    res.name = 'x %s'%scale
        elif np.size(scale)==2:      # x and y scales      
            res.x = scale[0] * res.x
            res.y = scale[1] * res.y
        else:
            raise ValueError('Multiply Data2D by wrong format!')
        return res

    def __rmul__(self,scale,*args,**kwargs):
        # this is called on the second term (self) if mul of the first (scale) failed in scale * self.
        
        #print('__rmul__ (%s,%s) calling %s__mul__ (%s)'%(2*(self,scale)))
        
        res = self.__mul__(scale,*args,**kwargs)
        #breakpoint()
        res.name = '%s x '%(scale)+(self.name if self.name else '') 
        return res
        
    def __matmul__(self,scale,*args,**kwargs):
        raise NotImplementedError ('Return a 2D surface (Data2D).')
        
    def __neg__(self):
        return self.__mul__(-1)

    def __rtruediv__(self,other):
        """division: accept Profile, scalars or 2-vector [x,y]. """
        # we are here because other/self didn't succeed with other.__truediv__
        # self is Profile, other can be integer or profile. 
        # it is result = other/self
        # print(self)
        # print(other)
        
        res = self.copy()
        if np.size(other) == 1:
            if isinstance(other,Profile):  # scalar
                sel = (self.y != 0) 
                res = self.copy()
                res.y[sel] = other*(1./self.y[sel])
                res.y[~sel] = np.nan
                #breakpoint()
                if other.units[1] == self.units[1]:
                    res.units[1] = ''
                else:
                    res.units[1]='%s / %s'%(other.units[1], self.units[1])
                res.name = '%s ratio'%(other.name if other.name is not None else '' ) + (' / %s'%(self.name) if self.name is not None else '') 
            else:     # value
                #breakpoint()
                res.y = other / res.y 
                if self.units[1]:
                    res.units[1] = '/'+self.units[1] 
                if self.name:
                    res.name = '%s / %s'%(other,self.name)
                else:
                    res.name = '%s / ratio'%(other)

        elif np.size(other)==2:      # x and y others    
            #return other.__truediv__(self)
            res.x = other[0] / res.x
            res.y = other[1] / res.y
        else:
            raise ValueError('__rtruediv__ Profile by wrong format!')
        return res

    def __truediv__(self,other):
    
        # __truediv__ is normally called, 
        # gets here for p2/p1 # p2/p1 = __truediv__(self=p2,other=p1)
        # gets here for p2/5 # p2/5 = __truediv__(self=p2,other=5)
        # 5/p2 will not call this, but will try to call 5.__truediv__, if it fails,
        #     p2.__rtruediv__ will be called instead

        '''https://stackoverflow.com/questions/37310077/python-rtruediv-does-not-work-as-i-expect
        
        __rtruediv__ only has priority over __truediv__ if the right-hand operand is an instance of a subclass of the left-hand operand's class.

        When you do 343 / x, NewInt is a subclass of int, so x's __rtruediv__ gets priority. When you do 343.3 / x, NewInt is not a subclass of float, so 343.3's __truediv__ gets priority.

        343.3.__truediv__(x) doesn't return NotImplemented, since float knows how to divide a float by an int. Thus, x.__rtruediv__ doesn't get called.
        '''
        # __rtruediv__ is implemented in Profile, so 1/other makes sense for Profile and numbers. 
        #breakpoint()
        res = self*(1./other)
        u = getattr(other,'units',['',''])[1]  # assegna units y or ''
        if self.units[1] == u:
            res.units[1] = ''
        else:
            res.units[1] = self.units[1]+('/'+u if u else '')
        if self.name is not None:
            #res.name = '__truediv__'
            res.name = '%s%s'%(self.name if self.name else "ratio",((' / %s'%getattr(other,'name',other)) if getattr(other,'name','') else ''))
        return res

    
    def merge(self,other,*args,**kwargs):
        
        if isinstance(other,Profile):
            res = merge_profiles([[self.x,self.y],[other.x,other.y]],*args,**kwargs)
            res = Profile(*res,units=self.units,name=self.name + " + " + other.name)
        else:
            raise ValueError("Unrecognized type in sum")
        return res
        
    def min (self):
        return np.nanmin(self.y)

    def max (self):
        return np.nanmax(self.y)    
        
    def sort (self,reverse=False):
        """return sorted copy."""
        res = self.copy()
        res.x,res.y= sort_profile(self.x,self.y,reverse=reverse)   
        return res
    
    def plot(self,title=None,*args,**kwargs):
        """plot profile using and setting automatically labels.
           Additional arguments are passed to `plt.plot`.
        
        Quite useless for profile, can be plot with `plt.plot(*P(),*args,**kwargs)"""
        #modeled over data2D.plot, non c'e' plot_profile
        from plotting.captions import legendbox
        from pyProfile.profile import get_stats
        
        stats=kwargs.pop('stats',0) #to change the default behavior
        loc=kwargs.pop('loc',0) #location for stats legend
        framealpha=kwargs.pop('framealpha',0.5) #transparency for stats legend
        l = kwargs.pop('label',None)
        if l is None and self.name:
            l = self.name
        res=plt.plot(self.x,self.y,label=l,*args,**kwargs)
        
        if stats: #add stats to plot
            legend=get_stats(self.x,self.y,units=self.units)
            l=legendbox(legend,loc=loc,framealpha=framealpha)
            
        plt.xlabel('X'+(" ("+self.units[0]+")" if self.units[0] else ""))
        plt.ylabel('Y'+(" ("+self.units[1]+")" if self.units[1] else ""))
  
        if title is None:
            if self.name is not None:
                title = self.name
        plt.title(title)
        
        return res
        
    plot=update_docstring(plot,plt.plot)


    def load(self,filename,*args,**kwargs):
        """A simple file loader using np.genfromtxt.
        Load columns from file in self.x and self.y."""
        #pdb.set_trace()
        self.x,self.y = np.genfromtxt(filename,unpack=True,*args,**kwargs)
    load=update_docstring(load,np.genfromtxt)

    def save(self,filename,*args,**kwargs):
        """Save data using `pyProfile.profile.save_profile`."""
        res = save_profile(filename,self.x,self.y,*args,**kwargs)
        return res
    save.__doc__=save_profile.__doc__
    

    def register(self,filename,*args,**kwargs):
        """Use pyProfile.profile.register_profile to rescale."""
        #pdb.set_trace()
        self.x,self.y = register_profile(x,y,*args,**kwargs)
    register=update_docstring(register,register_profile)



    '''
    from functools import update_wrapper
    #@update_wrapper(rotate_data)  #this doesn't work as I would like
    def rotate(self,angle,*args,**kwargs):
        """call data2D.rotate_data, which rotate array of an arbitrary angle in degrees in direction
        (from first to second axis)."""
        res = self.copy()
        res.data,res.x,res.y=rotate_data(self.data,self.x,self.y,angle,*args,**kwargs)
        return res
    rotate=update_docstring(rotate,rotate_data)
    '''

    '''
    def apply_to_data(self,func,*args,**kwargs):
        """apply a function from 2d array to 2d array to data."""
        res = self.copy()
        res.data=func(self.data,*args,**kwargs)
        return res
    '''
    
    def crop(self,*args,**kwargs):
        """crop profile making use of function profile.crop_data, where x,y are taken from self."""
        
        res=self.copy()
        res.x,res.y=crop_profile(self.x,self.y,*args,**kwargs)
        return res #
    crop=update_docstring(crop,crop_profile)
    
    def movingaverage(self,*args,**kwargs):
        """moving average using function profile.movingaverage, where x,y are taken from self."""
        
        res = self.copy()
        res.y = movingaverage(self.y,*args,**kwargs)
        return res #
    movingaverage = update_docstring(movingaverage,movingaverage)
        
    def rebin(self,*args,**kwargs):
        """rebin using function profile.rebin_profile, where x,y are taken from self."""
        
        res=self.copy()
        res.x,res.y=rebin_profile(self.x,self.y,*args,**kwargs)
        return res #
    
    rebin=update_docstring(rebin,rebin_profile)
    
    def level(self,degree=1,zero='mean',*args,**kwargs):
        """return a leveled profile calling profile.level_profile.
`zero option can be 'top', 'bottom' or 'mean', and is a facility to shift curves so that their min or max value is aligned to zero."""
        
        res=self.copy()
        res.x,res.y=level_profile(self.x,self.y,degree=degree,*args,**kwargs)
        if zero == 'top':
            res.y = res.y - np.nanmax(res.y)
        elif zero == 'bottom':
            res.y = res.y - np.nanmin(res.y)
        elif zero != 'mean':
            raise ValueError ("wrong value for `zero` option in level")
        return res
    level=update_docstring(level,level_profile)

    def resample(self,other,*args,**kwargs):
        """TODO, add option to pass x and y instead of other as an object."""
        res=self.copy()
        try:
            if self.units is not None and other.units is not None:
                if self.units[0] != other.units[0]:
                    raise ValueError('If units are defined they must match in Profile.resample.')
            res.x,res.y=resample_profile(*res(),*other(),*args,**kwargs)   
        except AttributeError: #assume other is an array
            res.x,res.y=resample_profile(*res(),None,other,*args,**kwargs)
        return res        
    resample=update_docstring(resample,resample_profile)


    def psd(self,wfun=None,rmsnorm=True,norm=1):
        """return a PSD object with psd of self. """

        f,p=profpsd(self.x,self.y,wfun=wfun,norm=norm,rmsnorm=rmsnorm)

        return PSD(p,self.x,f,units=self.units,name="")
    psd=update_docstring(psd,profpsd)

    def remove_nan_ends(self,*args,**kwargs):
        res = self.copy()
        res.x,res.y=profile.remove_nan_ends(self.x,self.y,*args,**kwargs)
        return res
    remove_nan_ends=update_docstring(remove_nan_ends,profile.remove_nan_ends)

    def std(self):
        """return standard deviation of data excluding nans"""
        return np.nanstd(self.y)

    def copy(self):
        """copy.deepcopy should work well."""
        return deepcopy(self)

    def printstats(self,label=None,fmt='%3.2g'):
        if label is not None:
            print(label)
        s=("%s PV: "+fmt+", rms: "+fmt)%(self.name,span(self.y,size=True),
                             np.nanstd(self.y))
        print(s)
        return s

    '''
    def align_interactive(self,other,find_transform=find_affine):
        """interactively set markers and align self to other.
        Alignment is performed using the transformation returned by
           find_transform(markers1,markers2) after markers are interactively set.
        Return aligned Data2D object.
        There is an experimental version for dlist in scripts."""
        from pySurf.scripts.dlist import add_markers
        m1,m2=add_markers([self,other])
        trans=find_transform(m1,m2)
        return self.apply_transform(trans)

    def remove_outliers(self,fill_value=np.nan,*args,**kwargs):
        """use dataIO.remove_outliers to remove outliers"""
        res=self.copy()
        m = remove_outliers(res.data,*args,**kwargs)
        res.data[~m] = fill_value
        return res
        
    remove_outliers=update_docstring(remove_outliers,dataIO.outliers.remove_outliers)

    def histostats(self,*args,**kwargs):
        res =data_histostats(self.data,self.x,self.y,units=self.units,*args,**kwargs)
        plt.title(self.name)
        return res
    histostats=update_docstring(histostats,data_histostats)

    def slope(self,*args,**kwargs):
        #import pdb
        #pdb.set_trace()
        scale=kwargs.pop('scale',None)
        if self.units is not None:
            if scale is None:
                if self.units[0]==self.units[1]:  #check if x and y in mm and z in um.
                    if self.units[0]=='mm' :
                        if self.units[0]=='mm' and self.units[2]=='um': scale=(1.,1.,1000.)
                else:
                    raise ValueError("x and y different units in slope calculation")
        else:
            scale=(1.,1.,1.)


        say,sax=slope_2D(self.data,self.x,self.y,scale=scale,*args,**kwargs)

        return Data2D(*sax,units=[self.units[0],self.units[1],'arcsec'],name=self.name + ' xslope'),Data2D(*say,units=[self.units[0],self.units[1],'arcsec'],name=self.name + ' yslope')
    slope=update_docstring(slope,slope_2D)
    '''

#from pySurf.psd2d import psd2d,plot_psd2d

class PSD(Profile):
    """It is a type of profile with customized behavoiur and additional properties
    and methods."""
    def __init__(self,*args,**kwargs):
        ''' super is called implicitly (?non vero)
        """needs to be initialized same way as Data2D"""
        #if a surface or a wdata,x,y are passed, these are interpreted as
        super().__init__(*args,**kwargs)
        '''
        super().__init__(*args,**kwargs)
                
    def plot(self,*args,**kwargs):
        u=kwargs.pop('units',self.units)
        return plot_psd(self.y,self.x,units=u,*args,**kwargs)
    
    def rms_power(self,plot=False,*args,**kwargs):
        """Calculate rms slice power as integral of psd. If plot is set also plot the whole thing."""
        
        raise NotImplementedError
        
        '''
        if plot:
            return plot_rms_power(self.y,self.data.self.x,units=self.units,*args,**kwargs)
        else:
            """this is obtained originally by calling rms_power, however the function deals with only scalar inmot for rms range.
            Part dealing with multiple ranges is in plot_rms_power, but should be moved to rms_power."""
            raise NotImplementedError
        '''

    def save(self,filename,*args,**kwargs):
        """Save psd."""
        self.save(filename,header='# f[%s] PSD[%s]'%self.units)

def load_from_blocks(file, *args, **kwargs):
    """read a list of profiles from a single file, extracting them using ``split_blocks``."""
    
    x,y = np.genfromtxt(file, unpack = True, *args, **kwargs)
    i = split_blocks(x)
    xx = split_on_indices(x,i)
    yy = split_on_indices(y,i)
    profiles = [Profile(x, y, name = '%03i'%i) for i,(x,y) in enumerate(zip(xx,yy))]
    return profiles
    
def load_plist(rfiles,reader=None,*args,**kwargs):
    """Read a set of profile file(s) to a plist. By default, files are split on blocks with spaces or changes in monotony.
    
    readers and additional arguments can be passed as scalars or lists,
    including arguments for ``load_from_blocks``.
    Return a Plist.
    
    2022/06/28 first implementation, copying from scripts/dlist.py, docstring not yet updated
    Mechanism for args and kwargs is quite basic and error-prone.

    You can pass additional arguments to the reader in different ways:
     - pass them individually, they will be used for all readers
         load_dlist(.. ,option1='a',option2=1000)
     - to have individual reader parameters pass them as dictionaries (a same number as rfiles),
         load_dlist(.. ,{option1='a',option2=1000},{option1='b',option3='c'},..)
         in this case reader must be explicitly passed (None is acceptable value for auto).

    Example:
        plist=load_plist(rfiles,reader=fitsWFS_reader,scale=(-1,-1,1),
                units=['mm','mm','um'])

        dlist2=load_dlist(rfiles,fitsWFS_reader,[{'scale':(-1,-1,1),
                'units':['mm','mm','um']},{'scale':(1,1,-1),
                'units':['mm','mm','um']},{'scale':(-1,-1,1),
                'units':['mm','mm','$\mu$m']}])
    """
    
    if isinstance(rfiles, str):
        profiles = load_from_blocks(rfiles, *args, **kwargs)
    else:
        if reader is None:
            #reader=auto_reader(rfiles[0])
            #reader = [auto_reader(r) for r in rfiles]
            reader = [None for r in rfiles]  # placeholder
            
        if np.size(reader) ==1:
            reader=[reader]*len(rfiles)
            
        ''' additional options see Dlist '''
        
        if kwargs : #passed explicit parameters for each reader
            # Note, there is ambiguity when rfiles and a kwargs value have same
            # number of elements()
            #pdb.set_trace()
            #vectorize all values
            for k,v in kwargs.items():
                if (np.size(v) == 1):
                    kwargs[k]=[v]*len(rfiles)    
                elif (len(v) != len(rfiles)):
                    kwargs[k]=[v]*len(rfiles)
                #else:  #non funziona perche' ovviamente anche chiamando esplicitamente, sara'
                #  sempre di lunghezza identica a rfiles.
                #    print ('WARNING: ambiguity detected, it is not possible to determine'+
                #    'if `%s` values are intended as n-element value or n values for each data.\n'+
                #    'To solve, call the function explicitly repeating the value.'%k)
        
        # 2020/07/10 args overwrite kwargs (try to avoid duplicates anyway).
        # args were ignored before.
        print(kwargs)
        if not args:  #assume is correct number of elements
            args = [[]]*len(rfiles)
        
        #pdb.set_trace()

        #transform vectorized kwargs in list of kwargs
        kwargs=[{k:v[i] for k,v in kwargs.items()} for i in np.arange(len(rfiles))]
        profiles = [Profile(file=wf1,reader=r,*a,**k) for 
            wf1,r,a,k in zip(rfiles,reader,args,kwargs)]
        
    self = Superlist()
    for p in profiles:    
        self.append(p)
        
    return self

'''
class Plist(Superlist):
    """A list of Profile objects on which unknown operations are performed serially.
    Useless, it is just a Superlist of profiles."""
    
    """
    def plot(self,*args,**kwargs):
        """plot over same plot each profile in the list."""
        
        for p in self:
            p.plot()
        
        return plt.gca()  
    """
    
    def __init__(self, files = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if files is not None:
            self = load_plist(files, *args, **kwargs)
'''
        
from dataIO.superlist import Superlist


class Plist(Superlist):

    def __init__(self, *args, **kwargs):
        
        if 'files' in kwargs:
            files = kwargs.pop('files')
            s = load_plist(files, *args, **kwargs)   # return a list
            super().__init__(s, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)        

    """ N.B.> could use linecollections for plotting https://matplotlib.org/stable/api/collections_api.html#matplotlib.collections.LineCollection
    """
    
def test_load_plist(rfiles = None):

    if rfiles is None:
        rfiles =[r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706.txt',
         r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706_sm11.txt',
         r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706_sm31.txt']

    pl = load_plist(rfiles,reader=None,units=['mm','um'],delimiter=',')
    #a.plot()
    
    print('\ntest method which perform actions (plot):')
    print(pl.plot())
    print('\ntest method which returns values (profile.min):')
    print(pl.min())
    print('\ntest chained method which returns Superlist (profile.crop):')
    pl.crop([2,3]).plot()
    
    """
    plist2=load_plist([],None,[{'scale':(-1,-1,1),
            'units':['mm','um']},{'scale':(1,1,-1),
            'units':['mm','um']},{'scale':(-1,-1,1),
            'units':['mm','$\\mu$m']}])
    return plist,plist2
    """
    
    return pl

def test_plist():

    rfiles =[r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706.txt', r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706_sm11.txt', r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\01_mandrel3_xscan_20140706_sm31.txt']
    
    p = Plist(files = rfiles, delimiter = ',')
    p.plot()
    plt.title('Plist initialized with files')
    
    plt.figure()
    profiles  = [ Profile(file = f, delimiter = ',') for f in rfiles]
    p = Plist(profiles)
    p.plot()
    plt.title('Initialized with Profiles read at init')

    plt.figure()
    profiles = load_plist(rfiles, delimiter = ',')
    p = Plist(profiles)
    p.plot()
    plt.title('Initialized with profiles loaded with function')
  
def test_class_init(wfile=None):
    """test init and plot"""
    from dataIO.fn_add_subfix import fn_add_subfix
    from pathlib import PureWindowsPath
    from pySurf.data2D import load_test,data
    
    x1,y1 = load_test_data(wfile,*args,**kwargs)

    plt.figure(1)
    plt.clf()
    plt.suptitle(relpath)
    plt.title('use plot_data function')
    plot_profile(x1,y1,aspect='equal')

    a=Profile(x1,y1)
    plt.figure(2)
    plt.clf()
    plt.title('From data')
    a.plot(aspect='equal')

    b=Profile(file=wfile,center=(0,0))
    plt.figure(3)
    plt.clf()
    plt.title('from filename')
    b.plot(aspect='equal')

    #b.save(os.path.join(outpath,os.path.basename(fn_add_subfix(relpath.as_posix(),"",".txt"))), makedirs=True)
    b.remove_nan_ends()
    plt.figure()
    plt.title('removed nans')
    b.plot(aspect='equal')
    
    
def test_profile_mul(p1,p2):
    #test mul
    print(p1)
    print(p1*p2) #calls p1.mul(p2) -> ok
    print(5*p1) #calls 5.mul(p2) -> falls back on p2.__rmul__
    print(p1*5) #calls p1.mul(5) -> ok
    
    #test mul
    print(span(p1.y))
    print('---------------')
    print(span((p1*p2).y)) #calls p1.mul(p2) -> ok
    print(span((5*p1).y)) #calls 5.mul(p2) -> falls back on p2.__rmul__
    print(span((p1*5).y)) #calls p1.mul(5) -> ok
    

def test_profile_mul2(p1,res):    

    test_plot(p1*res) #calls p1.mul(p2) -> ok
    test_plot(res*p1)
    test_plot(p1*p1)
    test_plot(p1*5) #calls p1.mul(p2) -> ok
    test_plot(5*p1)
    test_plot(5*res)
    test_plot(res*5) #calls p1.mul(5) -> ok

def test_profile_class():

    #test hardcore labels
    from pyProfile.profile_class import test_profile_mul2

    res=p1*2
    res.name = ''
    res.units = ['keV','']

    test_profile_mul2(p1,p2)
    test_profile_mul2(p1,res)
    test_profile_mul2(res,res)

    #test hardcore labels
    from pyProfile.profile_class import test_profile_div2

    res=p1*2
    res.name = ''
    res.units = ['keV','']

    test_profile_div2(p1,p2)
    test_profile_div2(p2,res)
    test_profile_div2(res,res)

def test_plot(p1):
    plt.figure()
    (p1).plot()
    plt.legend()
    print(p1)
    return p1


    
def test_profile_div(p1,p2):    
    
    test_profile_division(p1,p2)
    
    #test div
    print(span(p1.y))
    print('---------------')
    print(span((p1/p2).y)) #calls p1.div(p2) -> ok
    print(span((5/p1).y)) #calls 5.div(p2) -> falls back on p2.__rdiv__
    print(span((p1/5).y)) #calls p1.div(5) -> ok
    
    test_plot(p1)
    test_plot(p1/p2) #calls p1.mul(p2) -> ok
    test_plot(5/p1)
    test_plot(p1/5) #calls p1.mul(5) -> ok
 
def test_profile_div2(p1,res):    

    test_plot(p1/res) #calls p1.mul(p2) -> ok
    test_plot(res/p1)
    test_plot(p1/p1)
    test_plot(p1/5) #calls p1.mul(p2) -> ok
    test_plot(5/p1)
    test_plot(5/res)
    test_plot(res/5) #calls p1.mul(5) -> ok 
    

# test profile division
def test_profile_division(p1,p2):
    """ plot division of profile 2 by 1 and inverse of the inverse ratio based on np only and compare with division between objects, interpolate over 1st. Plot comparison with what is calculated by np only."""
    
    def divide (x1,y1,x2,y2):
        """division based only on np."""
        yy2=np.interp(x1,x2,y2)
        yr= y1/yy2
        return yr
    
    x1,y1 = p1()
    x2,y2 = p2()

    r = p2/p1
    plt.figure()
    
    r.plot()
    plt.plot(x2,divide(x2,y2,x1,y1),label='ratio p2/p1')
    plt.plot(x1,1/divide(x1,y1,x2,y2),label='1/ratio p1/p2')
    plt.legend()
    plt.xlim([0.1,14])
    plt.ylim([0.5,1.65])

if __name__=='__main__':
    plt.close('all')
    test_class_init()



import matplotlib.pyplot as plt
import numpy as np
from dataIO.span import span
from scipy.interpolate import interp1d
import copy
import pdb

def updating_plot(ax=None,title=None):
    """make the passed axis ax active (or the current one if not passed), such that
    it updates the color scale when zoomed to include only visible data.
    Axes is modified in place and returned.
    Usage:
        ax=updating_plot()  #make the current axis an updating_plot, axis is returned
        updating_plot(ax1)  #make the specific axis ax1 an updating_plot
    """
    
    """2017/03/09 BUG: if launched on a plot afteer already zoomed, gives runtime error:
    
    
    from plotting.updating_plot import updating_plot

    updating_plot()
    Out[9]: <matplotlib.axes._subplots.AxesSubplot at 0x41bbe908>
    C:\Anaconda2\lib\site-packages\numpy\lib\nanfunctions.py:675: RuntimeWarning: Mean of empty slice
      warnings.warn("Mean of empty slice", RuntimeWarning)
    C:\Anaconda2\lib\site-packages\numpy\lib\nanfunctions.py:1134: RuntimeWarning: invalid value encountered in less_equal
      isbad = (dof <= 0)
    C:\Anaconda2\lib\site-packages\numpy\lib\nanfunctions.py:1136: RuntimeWarning: Degrees of freedom <= 0 for slice.
      warnings.warn("Degrees of freedom <= 0 for slice.", RuntimeWarning)"""
      
    #note: it gives message is zoomed with both extremes out of range 
    # (zooming and panning near the edges). It fails calculating indices and cdata,
    # but still working
    # TODO: make class so that it is possible to enable/disable scale updates or disconnect the callbacks. Can be combined with different modes for setting scale crange (PV as now, std or frozen).
    # TODO: remove retrieval of figure using plt
    
    
    def on_lims_change(ax):
        nsigma=1.5
        
        ax.set_autoscale_on(False)
        #ax.figure.canvas.draw_idle(self)
        xl=ax.xaxis.get_view_interval()   #range of plot in data coordinates, sorted lr
        yl=ax.yaxis.get_view_interval()      #sorted bottom top
        #print xl,yl
        
        
        im=ax.images[-1]
        imdata=im.get_array()  #.filled(np.nan)  #this helps when a masked array is returned, for some reason it doesn't work
        dims=imdata.shape    #number of pixels
        #print dims
        extent=im.get_extent()      #this is the range for axis of  full data.
        #ixl=np.interp(xl,extent[0:2],[0,dims[1]])  #pixel indices for plotted data
        #iyl=np.interp(yl,extent[2:],[0,dims[0]])
        ixl=interp1d(extent[0:2],[0,dims[1]],assume_sorted=False,bounds_error=False)(xl).astype(int)  #pixel indices for plotted data
        iyl=interp1d(extent[2:],[0,dims[0]],assume_sorted=False,bounds_error=False)(yl).astype(int)
        #prevent empty array
        if np.abs(ixl[1]-ixl[0])<1:
            ixl[1]=ixl[0]+1
        if np.abs(iyl[1]-iyl[0])<1 : 
            iyl[1]=iyl[0]+1
        #print ixl[0],ixl[1],iyl[0],iyl[1]
        cdata=imdata[ixl[0]:ixl[1],iyl[0]:iyl[1]]
        #crange=np.nanmean(cdata)+[-np.nanstd(cdata),np.nanstd(cdata)] #this is a more sensitive view.
        if nsigma:
            crange=np.nanmean(cdata)+np.nanstd(cdata)*np.array([-1.,1.])*nsigma
        else:    
            crange=span(cdata)
        #print ax.__repr__(),cdata
        
        #make a copy of zoom history
        fig=ax.figure
        s = copy.copy( fig.canvas.toolbar._views )
        p = copy.copy( fig.canvas.toolbar._positions )
        
        title="min:%4.1f max:%4.1f rms:%4.3f"%(np.nanmin(cdata),np.nanmax(cdata),np.nanstd(cdata))
        #print crange
        ax.set_title(title)
        if not (im.colorbar is None):
            im.colorbar.remove()
        #if not (title is None):
        #    if title=="":
        #these commands resets colorbar
        plt.colorbar()
        plt.clim(crange)
        
        #restore zoom history
        fig.canvas.toolbar._views = s
        fig.canvas.toolbar._positions = p
        
        ax.figure.canvas.draw_idle()
    
    if ax is None: ax=plt.gca()
    
    """
    #failed attempt to remove existing on_lims_change if already existing. 
    
    if 'xlim_changed' in ax.callbacks.callbacks:
        try:
            ff=ax.callbacks.callbacks['xlim_changed']
            for k,f in ff.items():
                if f.func is on_lims_change:
                    print 'disconnect'
                    ax.callbacks.disconnect(k)
                else:
                    print 'no disco'
                    print f.func
                    print on_lims_change
        except KeyError:
            pass
    """
    
    ax.callbacks.connect('xlim_changed', on_lims_change)
    ax.callbacks.connect('ylim_changed', on_lims_change)
    return ax

def updating_mu(ax=None,title=None):
    """mock up. to solve nsigma and 
    make the passed axis ax active (or the current one if not passed), such that
    it updates the color scale when zoomed to include only visible data.
    Axes is modified in place and returned.
    Usage:
        ax=updating_plot()  #make the current axis an updating_plot, axis is returned
        updating_plot(ax1)  #make the specific axis ax1 an updating_plot
    """    
    
    def on_lims_change(ax,title=""):
        nsigma=1.5
        
        #calculate crange from plot, test values:
        crange=np.array([100.,1000.])*nsigma 
 
        ax.set_title(title+" nsigma: %4.1f"%(nsigma))
        ax.figure.canvas.draw_idle()
        
    if ax is None: ax=plt.gca()
    """
    #failed attempt to remove existing on_lims_change if already existing. 
    
    if 'xlim_changed' in ax.callbacks.callbacks:
        try:
            ff=ax.callbacks.callbacks['xlim_changed']
            for k,f in ff.items():
                if f.func is on_lims_change:
                    print 'disconnect'
                    ax.callbacks.disconnect(k)
                else:
                    print 'no disco'
                    print f.func
                    print on_lims_change
        except KeyError:
            pass
    """
    
    ax.callbacks.connect('xlim_changed', on_lims_change)
    ax.callbacks.connect('ylim_changed', on_lims_change)
    return ax

    
    
if __name__=="__main__":
    plt.ion()
    from pySurf.testSurfaces import test_makeGaussian
    a=test_makeGaussian(100)
    fig=plt.gcf()
    fig.canvas.set_window_title('A plot that updates colorscale on zoom')
    updating_plot()
    plt.show()


    